#########################################################################
# FILE:    Nb.tcl
# PURPOSE: Tcl7.4/Tk4.0 Discourse annotation tool
# AUTHOR:  Giovanni Flammia
# EMAIL:   flammia@goldilocks.lcs.mit.edu
# HTTP:    sls-www.lcs.mit.edu/flammia/Nb.html
# FTP:     sls-ftp.lcs.mit.edu/pub/flammia/Nb/
# DATE:    Summer 1995
# UPDATED: February 1996 - April 1996 - July 1996
#
# -------------------------------------------------------------------------
#        
# WISH LIST: Incorporate fully-compliant SGML parser and SGML I/O format
#            Add pretty printing capabilities
#            Windows 95, NT and Mac support
#            Include pcng_draw as a function call, not an exec...
#            More flexible editing of tags and boundaries
#            Coreference Links between tags
# 
# -------------------------------------------------------------------------
#
#            (N)[b] Copyright (c) 1995 - 1996
#            Spoken Language Systems Group
#            Laboratory for Computer Science
#            Massachusetts Institute of Technology
#
# -------------------------------------------------------------------------
#
#            Resources for Tcl and Tk
#
#            news:comp.lang.tcl
#            http://www.elf.org/tcltk-man-html.html
#            http://www.sco.com/Technology/tcl/Tcl.html
#            http://www.sco.com/IXI/of_interest/tcl/Tcl.html
#            http://web.cs.ualberta.ca/~wade/Auto/Tcl.html
#            ftp://ftp.cs.berkeley.edu/ucb/tcl
#
#
###########################################################################

# Note: I noticed some strange behavior of [$text compare $index1 <= $index2] when end-of-line
# indices are involved. Instead, I use [IndexToInt $index1] <= [IndexToInt $index2], where
# IndexToInt is defined in edit_functions.tcl

# check the version of Tcl and tk 
if [catch {set tkver $tk_version}] {
    puts "Sorry. The version of Tk you're running is too old for Nb."
    puts "Nb runs under Tk version 4.0 or later and Tcl version 7.4 or later."
    puts "You can download the latest version of Tcl/Tk at: http://www.sunlabs.com:80/research/tcl/"
    exit 
}


if [catch {set tclver [info tclversion]}] {
    puts "Sorry. The version of Tcl you're running is too old for Nb."
    puts "Nb runs under Tcl version 7.4 or later and Tk version 4.0 or later."
    puts "You can download the latest version of Tcl/Tk at: http://www.sunlabs.com:80/research/tcl/"
    exit 
}


if {$tkver < 4.0} {
    puts "Sorry. The version of Tk you're running is: $tkver."
    puts "Nb runs under Tk version 4.0 or later and Tcl version 7.4 or later."
    puts "You can download the latest version of Tcl/Tk at: http://www.sunlabs.com:80/research/tcl/"
    exit 
}


if {$tclver < 7.4} {
    puts "Sorry. The version of Tcl you're running is: $tclver."
    puts "Nb runs under Tk version 4.0 or later and Tcl version 7.4 or later."
    puts "You can download the latest version of Tcl/Tk at: http://www.sunlabs.com:80/research/tcl/"
    exit 
}

puts "Running Tcl version $tclver and Tk version $tkver"

puts "Loading: Nb.tcl ..."
# Process input arguments: 
# colormaps: see pages 285-287 of Brent Welch Book "practical programming in Tcl/Tk"
#            and "X Lib programming manual" volume 1 by Drian Nye for version 11
#            pages: 196-200

# Here is a list of six possible colormaps from the Nye book at page 193-198:
#
# read/write     read-only:        description:
# ----------------------------------------------------------
# greyscale      staticgray        monochrome/gray
# pseudocolor    staticcolor       single index for RGB
# directcolor    truecolor         decomposed index for RGB
# 
# Or: you can specify:
# "default" just to follow the X window resource


# Under windows, running from the wish console without arguments make it crash on exit, so...
if { $argc < 1 } {
   set argc 1
   set argv [list default]
}

if { $argc < 3 } {
    puts "Usage: Nb.tcl \[ \"colorMap \[Depth\]\" ConfigFile AnnotatorName InputFile \]"
}


for {set a 0} {$a < $argc} {set a [expr $a + 1]} {
    puts "Argument $a is: [lindex $argv $a]"
}


# define platform independent file manipulation functions: 
# native_to_unix, my_file_dirname, my_pwd, my_cd
proc native_to_unix {native} {

  switch -regexp -- $native {
    ^:$   {set this [pwd]}
    ^.$   {set this [pwd]}
    ^..$  {set this [file dirname [pwd]]}
    ^::$  {set this [file dirname [pwd]]}
    default {set this $native}
  }
  
  switch -regexp -- $this {
    ^[A-Za-z]\:[\/\\]
    { regsub -all -- {([\\]+)} $this "/" result }
    [:]+ 
    {
      if {[regexp {^[a-zA-Z]} $this match]} {
   set this "/$this"
      }
      regsub -all -- {([:]+)} $this "/" result
    }
    default {set result $this}
  }
  
  return $result
}


proc my_file_dirname {file_name} {
  return [native_to_unix [file dirname $file_name]]
}


proc my_pwd {} {
  return [native_to_unix [pwd]]
}


proc my_cd {directory_name} {
  cd [native_to_unix $directory_name]
}


# Begin Global variables:
set NB_SOURCE [my_file_dirname [info script]]
  
set sourceName ""
set sourceDir  ""
set clipBoard  [list]


# option and tagUnderCursor are visible at the bottom of the main window:
# Must be one of "Tag Words" "Tag Lines" "Edit Text" "Edit Signal" "View Signal":
set option "Tag Lines"
set tagUnderCursor ""
set idUnderCursor "                         "

# For rename Tag:
set entryTag ""
set oldTag   ""

# whiteTags lists all tags (segmentTags and commandTags) that refer to discourse segments,
# waveforms and pictures and therefore should not be considered parse tags to be included
# in bracketed text for View Parse.
set whiteTags [list]

# For file I/O:
set fileExists 0
set fileChanged 0
set sgmlFormat 0
set input_file_name ""
set output_file_name ""

# Config file with all the annotation parameters and Tag names
set colorMap      [lindex $argv 0]
if {[string compare $colorMap ""] == 0} {
   set colormap default
}

set configName    [lindex $argv 1]
if {[string compare $configName ""] == 0} {
   set configName $NB_SOURCE/config/Nb.config
}

set configTail    [file tail $configName]

# Name the program window
wm title . "Nb "
wm iconname . "Nb"
. configure -relief ridge  


set dot "."
if {![regexp {([a-zA-Z0-9]+)} $configTail match configPrefix]} {
    set configPrefix "."
} else {
    set configPrefix "$configPrefix$dot"
}


# external viewing and editing commands:
# Defaults: View Parse: cpng_draw.tcl "{bracket-ed sentence}"
#           View Segments: show_annotation.tcl file.sgml segmentTags
# Others:   choiceTable(Commands)


# user name for saving files
set annotatorName ""
set annotatorName [lindex $argv 2]
if {[string compare $annotatorName ""] == 0} {
   set annotatorName user
}

# why do I need these? 
set annotated_text_lines 25
set choices_text_lines   8

# Hashtables for cascade-menu Tags:
# fieldTable(firstLevelTag) = list with currently used index for firstLevelTag
set fieldTable(TopLevel)  [list]

# choiceTable(Tag) = list with pull-down menu options for Tag
set choiceTable(Options)  [list]
set choiceTable(TopLevel) [list]
set choiceTable(Colors)   [list]

# helpTable(Tag) = list of help lines for Tag to be displayed in the help window
set helpTable(TopLevel)   [list]

# formatTable(ChoiceTag) = A value associated with the choice Tag
# currently, if formatTable(TAG1) = "Tree" then apply  <TAG1> <TAG2> ... </TAG2> <TAG1>
set formatTable(TopLevel) ""

# priority table for nested tags like: <TAG1> <TAG2> ... </TAG2> <TAG1>
set priorityTable(TopLevel) 0

# For each index, keep a list of tags in preorder:
set beginAtIndex(0.0) [list]
set endAtIndex(0.0) [list]

# Each tag has the format topLevelTag_number and is associated with
# a name in idTable (a string), a color in colorTable, and a range.
# Hastables for text Tags:
set listOfSgmlTags [list]
set currentTags [list]
set cacheFrom(TopLevel)  0
set cacheTo(TopLevel)    0
set idTable(TopLevel)     "TopLevel"
set currentIndex(TopLevel) 1

# Recently used color in fieldTable(Colors)
set colorIndex 0
# set rangeTable(TopLevel)  { "1.1" "1.1" }
set rangeTable(TopLevel) [list]

# cacheTag(IdString)  returns the current Tag identified by idString
# cacheRange(Tag)     returns a list with the current range of text tagged with Tag
set cacheTag(TopLevelIdString) "TopLevel"
set cacheRange(TopLevel)       [list "1.1" "1.1"]
set speaker(0) ""
set segments(0) [list]

# Home directory for Nb tcl source files and global paths:
set slash "/"
set current_directory  [my_pwd]
# set current_directory  $current_directory$slash

puts "Working directory is: $current_directory"

# Global look parameters:"LightSlateGrey"
# default values:

set bg1                "Grey"
set fg1                "Midnightblue"
set bg2                "Grey"
set fg2                "Black"


# specialized widget values:
set text_bg            "DarkSlateBlue"
set text_fg            "White"
set default_color      "Grey"
set text2_bg           "Grey"
set text2_fg           "Black"
set scroll_bg          "Black"
set listbox_bg         "Grey"
set listbox_fg         "Black"

# Each tag has the format <topLevelTag>_<number> and is associated with
# a name in idTable (a string) and a color in colorTable.
# Hastables for text Tags:
set idTable(TopLevel)     "TopLevel"
set colorTable(TopLevel)  "Black"



# Load all the global parameters into the hashtables:
puts "Loading: file_functions.tcl..."
source $NB_SOURCE/file_functions.tcl
if {![file readable $configName]} {
    puts "Can't read configuration file: $configName. Assuming is: $NB_SOURCE/config/Nb.config"
    set configName $NB_SOURCE/config/Nb.config
    # exit
}

LoadConfigFile $configName fieldTable choiceTable helpTable 

proc SetSixtyFourColors {} {
  
  global choiceTable

  set rgb(0,0) 127
  set rgb(1,0) 100
  set rgb(2,0) 127

  set rgb(0,1) 100
  set rgb(1,1) 127
  set rgb(2,1) 100

  set rgb(0,2) 127
  set rgb(1,2) 127
  set rgb(2,2) 100
  
  set max_num_colors 64
  set this 0
  
  for { set i 0 } { $i < $max_num_colors } { incr i } {
    
    for { set j 0 } { $j < 3 } { incr j } {
      if { $j != $this } {
   set rgb($j,$this) [expr $rgb($j,$this) + [expr $j * 10]]
      }
      set rgb($this,$this) [expr $rgb($this,$this) + 4]
    }
    
    if { $rgb(0,$this) < 255 } {
      if { $rgb(1,$this) < 255 } {
   if { $rgb(2,$this) < 255 } {
     set this_color [format "#%2lx%2lx%2lx" $rgb(0,$this) $rgb(1,$this) $rgb(2,$this)]
     puts "Adding color: $this_color"
     lappend choiceTable(Colors) $this_color
    
   }
      }
    }

    incr this
    if { $this >= 3 } { set this 0 }
    
  }
}

# set global variables from configFile:
# set font1 -adobe-helvetica-bold-r-normal-*-20-100-*-*-*-*-iso8859-1
set small_font "-adobe-helvetica-bold-r-normal--12-120-75-75-p-70-iso8859-1"
set font1      "-adobe-helvetica-bold-r-normal--18-180-75-75-p-103-iso8859-1"

if {[info exists env(NB_FONT)]} {
  set font1 $env(NB_FONT)
  set text_font $env(NB_FONT)
} else {
  if {[info exists choiceTable(Font)]} {
    set text_font [lindex $choiceTable(Font) 0]
    set font1 [lindex $choiceTable(Font) 0]
    puts "Text Font: $text_font"
    set env(NB_FONT) $font1
  } else {
    set text_font "lucidasanstypewriter-bold-14"
  }
}
set font2 $font1

set maxLevelOfNesting 1
if {[info exists choiceTable(Nesting)]} {
  set maxLevelOfNesting [lindex $choiceTable(Nesting) 0]
}

set allowManyTags 0
if {[info exists choiceTable(AllowManyTags)]} {
  if {[lsearch -exact $choiceTable(AllowManyTags) "No"] != -1} {
    set allowManyTags 0
  } else {
    set allowManyTags 1
  }
}
set allowManyTags 0

puts "Allow Many tags: $allowManyTags (0 is no, 1 is yes)"


if {[info exists choiceTable(Colors)]} {
  set nbColors [llength $choiceTable(Colors)]
  puts "Colors: Loaded $nbColors color names from configuration file."
  
} else {

  SetSixtyFourColors
}

if {[info exists choiceTable(TopLevel)]} {
    puts "Tag Menus: $choiceTable(TopLevel)"
}

puts "Colors: $choiceTable(Colors)"





set textSelection ""
set rowSelection  ""

source $NB_SOURCE/connections.tcl
# End Global variables

# source $NB_SOURCE/show_segmentation.2.tcl
source $NB_SOURCE/display_segments.tcl




# Dialogue box functions:
if {$colorMap == "default" } {
    toplevel  .dbox -class frame -bg $bg1 -relief ridge
    puts "Using default color or grey scale map from the X window resource"
} else {
    if [catch {toplevel  .dbox -class frame -bg $bg1 -relief ridge -colormap new -visual $colorMap}] {
   puts "Nb cannot allocate the desired colormap: $colorMap"
   # exit
    }
}


wm  title .dbox       "Nb  Dialogue Window"
wm  iconname .dbox    "Nb  Dialogue Window"

frame .dbox.top -bg $bg1 
frame .dbox.mid -bg $bg1
frame .dbox.bot -bg $bg1

set dboxEntry   "dboxEntry"
set dboxMessage "dboxMessage"
set dboxOk      "Ok"
set dboxDismiss "Dismiss"



label .dbox.top.message -textvariable dboxMessage -font $font1 -fg $fg1 -bg $bg1
entry .dbox.mid.entry -textvariable dboxEntry  -font $font1 -relief sunken -bg $listbox_bg -fg $listbox_fg -width 62 -selectforeground $text_bg -selectbackground $text_fg 
.dbox.mid.entry config  
button .dbox.bot.ok -textvariable dboxOk -font $font1 -command "dboxApply" -relief raised -bg $bg1 -fg $fg1 -padx 20 
button .dbox.bot.dismiss -textvariable dboxDismiss -font $font1 -command "dboxWithdraw" -relief raised -bg $bg1 -fg $fg1 -padx 20 

pack .dbox.top .dbox.mid .dbox.bot -side top -expand 1 -fill x
pack .dbox.top.message -side top
pack .dbox.mid.entry -side top 
pack .dbox.bot.ok -side left
pack .dbox.bot.dismiss -side right 

bind .dbox.mid.entry <Return> {
    dboxApply
}

proc dboxWithdraw {} {
    global dboxMessage dboxEntry dboxOk dboxDismiss
    global dboxVariable textArea fieldTable colorTable idTable
    set dboxOk "Ok"
    set dboxDismiss "Dismiss"
    switch -regexp $dboxMessage {
      {Save changes} {
        exit 0
      }
      {email address} {
        exit 0
      }
    }
    catch {wm withdraw .dbox}
}


# When dboxApply is called, the text selection might be lost, 
# but its copy should be saved in ViewTags.copy:
proc dboxApply {} {

  global dboxMessage dboxEntry dboxVariable textArea fieldTable colorTable idTable
  global fileValue fileLabel fileCommand fileFunction annotatorName

  switch -regexp $dboxMessage {
    {Name} {
      if {[regexp {([-_a-zA-Z0-9]+)} $dboxVariable match firstTag]} {
    if {[$textArea.text tag ranges ViewTags.copy] != ""} {
          regexp   {([-_a-zA-Z0-9]+)} $dboxVariable match firstTag
          set tag  "$firstTag $dboxEntry" 
          TagNow $tag
          FindAllTags
          killSegmentationWindow
    } else {
          WarningMessage "I lost the text selection for your tag. Please Retry."
    }
      }
      dboxWithdraw
    }
    {Save changes} {
      set fileValue   $dboxEntry
      set fileLabel   "Save File"
      set fileCommand ""
      set fileFunction "Save"
      SaveFile
      exit 0
    }
    {Find} {
      FindText $textArea.text $dboxEntry
    }
    {email address} {
      if {[regexp {^[.-_a-zA-Z0-9]+@[-_a-zA-Z0-9]+\.} $dboxEntry]} {
        set annotatorName $dboxEntry
        puts "Setting annotator name to: $annotatorName"
        catch {wm withdraw .dbox}
        WarningMessage "I will contact you at: $annotatorName to send you the gift certificates!"
      } else {
        WarningMessage "$annotatorName does not seem to be an email address. I am expecting
something like: user@company.com, user@school.edu..."
      }
    }
  }
}

# Name the window and global variables:
toplevel .view -class frame -bg $bg1 -relief raised -colormap .dbox -visual $colorMap
wm title .view "Nb  View Annotation"
wm iconname .view "Nb  View Annotation"

toplevel .outline -class frame -bg $bg1 -relief raised -colormap .dbox -visual $colorMap
wm title .outline "Nb  Annotation Outline"
wm iconname .outline  "Nb  Annotation Outline"

# Name the window and global variables:
toplevel .colors -class frame -bg $bg1 -relief raised -colormap .dbox -visual $colorMap
wm title .colors "Nb  Segment List"
wm iconname .colors "Nb  Segment List"

source $NB_SOURCE/view_color_palette.tcl


# File Selection Pop-Up window is a listbox called ".d":
# create a listbox:
toplevel  .chwindow -class frame -bg $bg1 -relief ridge -colormap .dbox -visual $colorMap
wm  title .chwindow       "Nb  Edit Window"
wm  iconname .chwindow    "Nb  Edit Window"



set typeArea         .chwindow.1
set choicesArea      .chwindow.2
set bchoicesArea     .chwindow.3

frame $typeArea      -bg  $bg1 
frame $choicesArea   -bg  $bg1
frame $bchoicesArea  -bg  $bg1

pack $choicesArea  -side top -fill x -expand true
pack $typeArea     -side top -fill x -expand true
pack $bchoicesArea -side bottom -fill x -expand true


# Frames from top to bottom and from left to right:
set main_window      .nb
set menuArea         .nb.1       
set textArea         .nb.4 
set bottomArea       .nb.5



# main window area
frame $main_window      -bg  $bg1 -relief raised -colormap .dbox -visual $colorMap
frame $menuArea         -bg  $bg1 -relief raised -bd 2  -colormap .dbox -visual $colorMap
frame $textArea         -bg  $bg1 -colormap .dbox -visual $colorMap
frame $bottomArea       -bg  $bg1 -colormap .dbox -visual $colorMap
frame $bottomArea.1     -bg  $bg1 -colormap .dbox  -visual $colorMap
frame $bottomArea.2     -bg  $bg1 -colormap .dbox -visual $colorMap


# allow interactive resizing of main window between minsize and maxsize
# wm minsize .  100 40
# wm maxsize .  120 60
# wm aspect  .  1 4 4 1


# menu bar sits on top side of Nb main window
menubutton $menuArea.file   -text "File"  -padx 10   -font $font1 -menu $menuArea.file.menu -bg  $bg1 -fg $fg1
   

source $NB_SOURCE/output_functions.tcl


source $NB_SOURCE/edit_functions.tcl
source $NB_SOURCE/edit_menu_handlers.tcl
menubutton $menuArea.mode   -text "Mode"  -padx 10 -font $font1   -menu $menuArea.mode.menu -bg  $bg1 -fg $fg1 
 
label $menuArea.optionValue -textvariable option -font $font1 -fg "DarkSlateBlue" -bg $bg1

pack $main_window -expand 1 -fill both
# pack $choicesArea -expand 0 
pack $menuArea    -side top -fill x
# pack $typeArea    -side top -fill x
# pack $choicesArea -side top -fill x


# File menu
set this_menu $menuArea.file.menu
menu $this_menu -bg  $bg1 -tearoff false 


# This one is not so good in Windows if you want to avoid exec...
# $this_menu add command -label "New"      -font $font1 \
#    -command "exec $NB_SOURCE/Nb $configName >& /dev/null &"


# set start_new_nb "exec wish4.1 -f bin/Nb.tcl default bin/Test.config $annotatorName Table-Of-Contents &"

# $this_menu add command -label "New Nb Window"  -font $font1 -command $start_new_nb 

$this_menu add command -label "Open"      -font $font1 \
    -command { 
      # .d.l.dir_value view [expr [.d.l.dir_value index end] - 80]; 
      FirstSaveFile
      set fileLabel   "Find"
      set fileValue   ""
      set fileCommand ""
      wm deiconify .d 
      set fileFunction "Open"
      DisplayFiles .d.m.directory $current_directory
      .d.m.directory configure -foreground black
      wm iconname .d "Nb  Open File"
      wm title .d "Nb  Open File"
      raise .d .
    }


$this_menu add command -label "Save    " -font $font1 \
    -command {
      set fileChanged 1
      set fileExists 1
      set fileLabel   "Save File"
      set fileValue   ""
      set fileCommand ""
      set fileFunction "Save"
      if {$output_file_name != ""} {
        set fileValue [file tail $output_file_name]
      } else {
        if {$input_file_name != "" } {
            set suffix ".nb" 
            set dot    "."
            set fileValue [file tail $input_file_name]
        }
      }
      wm deiconify .d
      DisplayFiles .d.m.directory $current_directory
      wm iconname .d "Nb  Save File"
      wm title .d "Nb  Save File"
      raise .d .
      .d.m.directory configure -foreground white
    }



proc QuitProgram {} {
    global fileExists fileChanged
    puts "Called QuitProgram\n"
    if $fileExists {
      if $fileChanged {
         catch {MayBeSaveFile}
      } else {
         killSegmentationWindow
         catch {wm destroy .}
         exit 0
      }
    } else {
      killSegmentationWindow
      catch {wm destroy .}
      exit 0
    }
}

$this_menu add command -label "Exit   " -font $font1 -command { $textArea.text config -cursor watch ; QuitProgram }

# Edit menu
set this_menu $menuArea.mode.menu
menu $this_menu -bg  $bg1  -tearoff false

if {[lsearch -exact $choiceTable(Options) "TagSelection"] != -1} {
  $this_menu add radiobutton -label "Tag Selection "  -font $font1 -variable option -value "Tag Selection" -command {$textArea.text configure -state disabled ; ReBind }
}

if {[lsearch -exact $choiceTable(Options) "TagWords"] != -1} {
  set option "Tag Words"
  $this_menu add radiobutton -label "Tag Words     "  -font $font1 -variable option -value "Tag Words" -command {$textArea.text configure -state disabled ; ReBind }
}

if {[lsearch -exact $choiceTable(Options) "TagLines"] != -1} {
  set option "Tag Lines"
  $this_menu add radiobutton -label "Tag Lines     "  -font $font1 -variable option -value "Tag Lines" -command {$textArea.text configure -state disabled ; ReBind }
}

if {[lsearch -exact $choiceTable(Options) "EditText"] != -1} {
  $this_menu add radiobutton -label "Edit Text     "  -font $font1 -variable option -value "Edit Text" -command {$textArea.text configure -state normal ; ReBind } 
}

# Set font menu
menubutton $menuArea.font   -text "Font"  -padx 10 -font $font1   -menu $menuArea.font.menu -bg  $bg1 -fg $fg1 
 
set this_menu $menuArea.font.menu
menu $this_menu -bg  $bg1  -tearoff false

$this_menu add radiobutton -label "Fixed"  -font $font1 -command {$textArea.text configure -font "fixed"}
$this_menu add radiobutton -label "Helvetica Small"  -font $font1 -command {$textArea.text configure -font "-adobe-helvetica-bold-r-normal--12-120-75-75-p-70-iso8859-1"}
$this_menu add radiobutton -label "Helvetica Medium"  -font $font1 -command {$textArea.text configure -font "-adobe-helvetica-bold-r-normal--14-100-100-100-p-82-iso8859-1"}
$this_menu add radiobutton -label "Helvetica Large"  -font $font1 -command {$textArea.text configure -font "-adobe-helvetica-bold-r-normal--18-180-75-75-p-103-iso8859-1"}
$this_menu add radiobutton -label "Courier Small"  -font $font1 -command {$textArea.text configure -font "-adobe-courier-bold-r-normal--12-120-75-75-m-70-iso8859-1"}
$this_menu add radiobutton -label "Courier Medium"  -font $font1 -command {$textArea.text configure -font "-adobe-courier-bold-r-normal--14-100-100-100-m-90-iso8859-1"}
$this_menu add radiobutton -label "Courier Large"  -font $font1 -command {$textArea.text configure -font "-adobe-courier-bold-r-normal--18-180-75-75-m-110-iso8859-1"}
$this_menu add radiobutton -label "Lucida Small"  -font $font1 -command {$textArea.text configure -font "-b&h-lucida-bold-r-normal-sans-12-120-75-75-p-79-iso8859-1"}
$this_menu add radiobutton -label "Lucida Medium"  -font $font1 -command {$textArea.text configure -font "-b&h-lucida-bold-r-normal-sans-14-140-75-75-p-92-iso8859-1"}
$this_menu add radiobutton -label "Lucida Large"  -font $font1 -command {$textArea.text configure -font "-b&h-lucida-bold-r-normal-sans-18-180-75-75-p-120-iso8859-1"}
$this_menu add radiobutton -label "Times Small"  -font $font1 -command {$textArea.text configure -font "-adobe-times-bold-r-normal--12-120-75-75-p-67-iso8859-1"}
$this_menu add radiobutton -label "Times Medium"  -font $font1 -command {$textArea.text configure -font "-adobe-times-bold-r-normal--14-140-75-75-p-77-iso8859-1"}
$this_menu add radiobutton -label "Times Large"  -font $font1 -command {$textArea.text configure -font "-adobe-times-bold-r-normal--18-180-75-75-p-99-iso8859-1"}


# default annotated text window and its scroll bar
# lay out at right area of main text window:
pack $textArea -side top -expand 1 -fill both

frame $textArea.x -bg $bg1 -colormap .dbox -visual $colorMap
frame $textArea.y -bg $bg1 -colormap .dbox -visual $colorMap

# Main window text widget: by default you can't write on it:
text      $textArea.text   -relief sunken -font $text_font -bg $text_bg -fg $text_fg \
    -yscrollcommand "$textArea.scroll set" \
    -xscrollcommand "$textArea.x.scroll set" \
    -state normal \
    -selectbackground $text_fg \
    -selectforeground $text_bg \
    -insertbackground HotPink \
    -height $annotated_text_lines -wrap none -setgrid 1 
scrollbar $textArea.scroll  -orient vertical -command "$textArea.text yview" -relief sunken -bg $scroll_bg
scrollbar $textArea.x.scroll -orient horizontal -command "$textArea.text xview" -relief sunken -bg $scroll_bg
$textArea.text config -insertbackground yellow -selectforeground Black


# Keyboard and mouse event bindings:
set cursor_position(x) 0
set cursor_position(y) 0
set found_tags 0

# Unfortunately, this one is not working for MacOS Tk 4.2
# The listbox is not showing up most of the times...
proc text_event_handler {x y} {
  
  global cursor_position .chwindow textArea found_tags sleep_now
  global cacheRange cacheTag 
  global whiteAndNbTags typeArea tagUnderCursor idUnderCursor idTable choicesArea currentTags currentIndex
  global editingAction editingMessage renameMessage entryTag oldTag bchoicesArea
  global selectedId oldTag fileExists entryTag replaceTag cursor_position

  set sleep_now 1

  set cursor_position(x) $x
  set cursor_position(y) $y
  catch {wm withdraw .chwindow}
  set y_position [expr $cursor_position(y) + 15]
  set x_position [expr $cursor_position(x) - 15]

  wm geometry  .chwindow "+$x_position+$y_position"
  wm title     .chwindow "Nb Edit Tags"
  wm deiconify .chwindow
  raise .chwindow .
  
  if {[$textArea.text tag ranges ViewTags.sel] != ""} {
    $textArea.text tag delete ViewTags.sel
  }
  if {[$textArea.text tag ranges ViewTags.copy] != ""} {
    $textArea.text tag delete ViewTags.copy
  }
  selection clear $textArea.text
  set found_tags 0
  EditTagsUnderCursor
  
  set sleep_now 0
  update idletasks
  
}

# handles mouses with one, two, or three buttons:
bind $textArea.text <ButtonPress-2> {text_event_handler %X %Y}
bind $textArea.text <ButtonPress-3> {text_event_handler %X %Y}


set sleep_now 0
bind $textArea.text <Motion> {
  if {$sleep_now == 0} {
    set tagUnderCursor ""
    set idUnderCursor "                        "
    ListTagsUnderCursor
  }
}







set binding "after"

# Clear ViewTags.sel every time the user uses the mouse:
bind $textArea.text <ButtonPress-1> {
  set cascadeTag ""
  if {[$textArea.text tag ranges ViewTags.sel] != ""} {
    $textArea.text tag delete ViewTags.sel
  }
  if {[$textArea.text tag ranges ViewTags.copy] != ""} {
    $textArea.text tag delete ViewTags.copy
  }
  focus $textArea.text
}

set hyper 0

# fake "sel" tag with "ViewTags.sel" tag, and expand user selection to Tag Words or Tag Lines:
bind $textArea.text <ButtonRelease-1> {
  set range [$textArea.text tag ranges sel]
  if {$hyper != 1} {
    if {$range != ""} {
      set fromIndex [lindex $range 0]
      set toIndex   [lindex $range 1]
      regexp {^([0-9]+)\.([0-9]+)} $toIndex match selRow selCol
      switch $option {
    "Tag Words" 
    {
      set fromIndex [$textArea.text index "$fromIndex wordstart"] 
      set toIndex   [$textArea.text index "$toIndex wordend"]
      regexp {^([0-9]+)\.([0-9]+)} $toIndex match rowEnd colEnd
      if { $rowEnd > $selRow } {
        set toIndex [lindex $range 1]
      }
      if { $selCol == 0} {
        set toIndex [lindex $range 1]
      }
      $textArea.text tag add ViewTags.copy $fromIndex $toIndex
      $textArea.text tag configure ViewTags.copy -background $text_fg -foreground $text_bg
    } 
    "Tag Lines" 
    {
      set fromIndex [$textArea.text index "$fromIndex linestart"]
      set toIndex   [$textArea.text index "$toIndex lineend"]
      regexp {^([0-9]+)\.([0-9]+)} $toIndex match rowEnd colEnd
      if { $rowEnd > $selRow } {
        set toIndex [lindex $range 1]
      }
      if { $selCol == 0 } {
        set toIndex [lindex $range 1]
      }
      $textArea.text tag add ViewTags.copy $fromIndex $toIndex
      $textArea.text tag configure ViewTags.copy -background $text_fg -foreground $text_bg
    }
    default
    {
      $textArea.text tag add ViewTags.copy $fromIndex $toIndex
      $textArea.text tag configure ViewTags.copy -background $text_fg -foreground $text_bg
    }
      }
    }
    focus $textArea.text
    set hyper 0
  } else {
    set hyper 0
  }
}




set xpad [expr [$textArea.scroll cget -width] + 2* \
         ([$textArea.scroll cget -bd] + \
         [$textArea.scroll cget -highlightthickness])]


frame $textArea.x.pad -width $xpad -height $xpad -bg $bg1 -colormap .dbox -visual $colorMap

pack      $textArea.x -side bottom -fill x
pack      $textArea.x.pad -side right
pack      $textArea.x.scroll -side left -fill x -expand true

pack      $textArea.y -side top -fill both -expand true
pack      $textArea.scroll -in $textArea.y  -side right -fill y 
pack      $textArea.text   -in $textArea.y  -side left -fill both -expand 1 



# Pack bottomArea control values:
set statusMessage "Ok"

label $bottomArea.tagValue -anchor sw -textvariable tagUnderCursor -font $font1 -fg black -bg $bg1 -justify left -pady 5 -borderwidth 5
label $bottomArea.status -anchor sw -textvariable statusMessage -font $font1 -fg darkSlateBlue -bg $bg1 -justify right -pady 5 -borderwidth 5

pack  $bottomArea -side bottom -fill x
pack  $bottomArea.tagValue -side left
pack  $bottomArea.status -side right


# Dialogue box functions:
toplevel  .id -class frame -bg $bg1 -relief ridge -colormap .dbox -visual $colorMap
wm  geometry .id 500x100
wm  title .id       "Nb  Tags Under Cursor"
wm  iconname .id    "Nb  Tags Under Cursor"
wm  minsize .id 500 100

frame .id.frame -bg $bg1 -width 400 -height 100
label .id.frame.label -anchor sw -textvariable idUnderCursor -font $font1 -fg black -bg $bg1 -justify left -padx 5 -pady 5 -borderwidth 5 

pack .id.frame -side left -expand true -fill both
pack .id.frame.label -side left 



set currentTag ""
set selectedTag ""
set tagTable(NbDefaultTag) ""
set tearoff true


# Recursive procedure for creating Tag pull-down menus:
proc AddMenu {thisTag thisMenu tagSequence selFont selBg} {

  global fieldTable cascadeTag choiceTable formatTable menuWidgets tearoff colorForId
  set fieldTable($thisTag) [list 0]
  set tagValue "$tagSequence $thisTag"

  # More choices for this Tag: create pull-down menu:
  if {[info exists choiceTable($thisTag)]} {
    puts "Adding Pull-down menu: $thisTag to: $thisMenu"
    $thisMenu add cascade -label $thisTag -font $selFont -menu $thisMenu.m_$thisTag 
    menu $thisMenu.m_$thisTag -bg $selBg -tearoff $tearoff
    set w $thisMenu.m_$thisTag
    set v $tagValue
    # Is thisTag invisible (i.e. thisTag is just a choice for another menu):
    if {[info exists formatTable($thisTag)]} {
      if {$formatTable($thisTag) == "Invisible"} {
        set v "$tagSequence"
        set tagValue "$tagSequence"
      } 
    }
    lappend menuWidgets($thisTag) $w $v
    # Add 
    # exit radiobutton:
    #if {$tagValue != $tagSequence} {
    #  $thisMenu.m_$thisTag add radiobutton -label "."  -font $selFont -variable cascadeTag -value $tagValue -command {
    #    PleaseTag "$cascadeTag" 
    #  }
    #}
    if {[regexp {(^$thisTag | $thisTag | $thisTag$)} $tagSequence match value]} {
      puts "Warning: Skipping pull-down menu for $thisTag (detected infinite loop?)."
      exit
    } else {
      set doneList [list]
      # Add all the choices:
      foreach secondTag $choiceTable($thisTag) {
        if {[lsearch -exact $doneList $secondTag] == -1} {
          AddMenu $secondTag $thisMenu.m_$thisTag $tagValue $selFont $selBg
          lappend doneList $secondTag
        }
      }
    }
  } else {
    # puts "Adding radio-button: $thisTag"
    # Terminal choice: create radiobutton:
    puts "Looking for color for: \"$tagValue\""
    if {[regexp {^[ ]*([-\_a-zA-Z0-9]+)[ ]+(.+)$} $tagValue match head name]} {
      if {[info exists colorForId($head.$name)]} {
        set tagColor $colorForId($head.$name)
      } else {
        set tagColor white
      }
    } else {
      set tagColor white
    }
    $thisMenu add radiobutton -label $thisTag -font $selFont -background DarkSlateBlue -foreground $tagColor -variable cascadeTag -value $tagValue -command {
      PleaseTag "$cascadeTag"
    }
  }
}


# Load the keyboard bindings specific to Nb events:
source $NB_SOURCE/bindings.tcl 



### label $typeArea.label -textvariable renameMessage -font $font1 -bg $bg1 -fg $fg1 -pady 10

menubutton $typeArea.label -textvariable renameMessage -font $font1 -bg $bg1 -fg $fg1 -menu $typeArea.label.menu -bg $bg1 -fg $fg1
menu $typeArea.label.menu -fg $fg1 -bg $bg1 -tearoff false
set topMenu $typeArea.label.menu
set topSequence ""
set doneList [list]
set tearoff false
foreach topTag $choiceTable(TopLevel) {
    if {[lsearch -exact $doneList $topTag] == -1} {
      AddMenu $topTag $topMenu $topSequence $font1 $bg1
      lappend doneList $topTag
    }
}


entry $typeArea.value -textvariable entryTag -selectbackground $listbox_fg -selectforeground $listbox_bg \
    -font $font1 -relief sunken -bg $listbox_bg -fg $listbox_fg -width 32
$typeArea.value config  -insertbackground yellow




# this is the  running text
# listbox with the current Tags listing:
frame     $choicesArea.1 -bg $bg1


label     $choicesArea.1.title -textvariable editingAction -font $font1 -bg $bg1 -fg $fg1


label     $choicesArea.1.label -textvariable editingMessage -font $font1 -bg $bg1 -fg $fg1

listbox   $choicesArea.directory -relief raised -yscrollcommand "$choicesArea.scroll_directory set" \
    -exportselection 0 -bg $listbox_bg -fg $listbox_fg -font $font1 -width 32 -height 2 -selectbackground white -selectforeground DarkSlateBlue -highlightcolor white
scrollbar $choicesArea.scroll_directory -command "$choicesArea.directory yview" -bg $scroll_bg -relief sunken

frame     $typeArea.pad -bg $bg1 -width 30 -height 20

pack      $choicesArea.1 -side top
pack      $choicesArea.1.title $choicesArea.1.label -side top 
pack      $choicesArea.scroll_directory -side right -fill y 
pack      $choicesArea.directory -side top -fill both -expand 1 
pack      $typeArea.label -side top
pack      $typeArea.value -side top -fill x -expand 1
pack      $typeArea.pad -side top -fill x -expand 1


button $bchoicesArea.delete -text "Delete Tag" -font $font1 -command "DeleteThisTag ; wm withdraw .chwindow" -relief raised -bg $bg1 -fg $fg1 -padx 20 

# Add here: it's a pull down menu...

button $bchoicesArea.rename -text "Change Tag" -font $font1 -command "NameThisTag" -relief raised -bg $bg1 -fg $fg1 -padx 20 
button $bchoicesArea.add -text "Add To Menu" -font $font1 -command "AddThisTag" -relief raised -bg $bg1 -fg $fg1 -padx 20 
button $bchoicesArea.dismiss -text "Dismiss" -font $font1 -command {wm withdraw .chwindow ; set replaceTag 0} -relief raised -bg $bg1 -fg $fg1 -padx 20 


pack   $bchoicesArea.dismiss -side right
### pack   $bchoicesArea.delete $bchoicesArea.rename $bchoicesArea.add  -side left
pack   $bchoicesArea.delete $bchoicesArea.add  -side left


menubutton $menuArea.tag -text "Tag" -padx 10   -font $font1 -menu $menuArea.tag.menu -bg $bg1 -fg $fg1 
  
menubutton $menuArea.edit_tag -text "Edit" -padx 10  -font $font1 -menu $menuArea.edit_tag.menu -bg $bg1 -fg $fg1 

menubutton $menuArea.view -text "View" -padx 10 -font $font1 -menu $menuArea.view.menu -bg $bg1 -fg $fg1 
menu $menuArea.view.menu -bg $bg1 -tearoff false

source $NB_SOURCE/help_functions.tcl
menubutton $menuArea.help -text "Help" -padx 10  -font $font1 -menu $menuArea.help.menu -bg $bg1 -fg $fg1 -justify right

menu $menuArea.help.menu -bg $bg1 -tearoff false

set selectionFrom "nil"
set selectionTo   "nil"

# Customize Tag menu as a recursive pull-down 
# cascaded menu according to TopLevel keywords defined in choiceTable(TopLevel)

set this_menu $menuArea
menu $this_menu.tag.menu -font $font1 -bg $bg1 -tearoff false
set menuWidgets(Tag) [list $this_menu.tag.menu "" $typeArea.label.menu ""]
# puts "Top Level tags: $formatTable(TopLevel)"
set cascadeTag ""


proc AddCommandToMenus { item menuList thisTag } {
  global font1 cascadeTag
  puts "Entering: AddCommandToMenus: $item $menuList $thisTag"
  for {set i 0} {$i < [llength $menuList]} {set i [expr $i + 2]} {
    set thisMenu    [lindex $menuList $i]
    set tagSequence [lindex $menuList [expr $i + 1]]
    if {[string length $thisMenu] > 1} {
      $thisMenu add radiobutton -label $thisTag -font $font1 -variable cascadeTag -value "$tagSequence $thisTag"  -command {
   PleaseTag "$cascadeTag" 
      }
    }
  }
  if {[llength $menuList] >= 2} {
    WarningMessage "Added new choice: \"$thisTag\" to all menus called \"$item\""
  }
}




set dboxVariable ""
proc PromptForFeatures {tag} {
  global dboxMessage dboxEntry dboxVariable
  set dboxVariable  $tag
  set dboxEntry    ""
  set dboxMessage  "Name for New $tag:"
  wm  deiconify .dbox
  raise .dbox .
}



# This one gets called when the user clicks on the tag pull-down menu:
proc PleaseTag {tag} {
  global beginAtIndex endAtIndex formatTable choiceTable textArea editingAction
  global replaceTag
  if {[$textArea.text tag ranges sel] != ""} {
    if {[regexp {^[ ]*([-a-zA-Z0-9]+)} $tag match firstTag] == 1} {
      if {[info exists formatTable($firstTag)]} {
   if {$formatTable($firstTag) == "Features"} {
     if {[regexp {^[ ]*([-a-zA-Z0-9]+)[ ]+([-a-zA-Z0-9]+)} $tag match first rest] == 1} {
       StartTag "$tag"
     } else {
       PromptForFeatures "$tag"
     }
   } else {
     StartTag "$tag"
   }
      } else {
   StartTag "$tag"
      }
    } else {
      WarningMessage "PleaseTag: $tag no first tag match? replaceTag is $replaceTag"
    } 
  } else {
    if {$replaceTag == 1} {
      StartTag "$tag"
    }
  }
}



# Here: Loop for all keywords Tag in tagSequence to see whether
# it's <Tag="Rest"> ... </Tag> or (if formatTable(Tag) = Features)
# <Tag> <Rest> ... </Rest> </Tag> (if formatTable(Tag) = Tree)
proc StartTag {tagSequence} {
  
  global textArea fieldTable clipBoard formatTable colorTable idTable choiceTable
  global replaceTag option
  
  
  
  if {[string compare $option "Edit Text"] != 0} {
    
    set line $tagSequence
    puts "Entering: StartTag $tagSequence"
    
    if {[regexp {[a-zA-Z0-9]+} $line]} {
      if {[$textArea.text tag ranges sel] != ""} {
	set pred 1
      } else {
	set pred 0
      }
      if {$replaceTag == 1} {
	incr pred
      }  
      if { $pred > 0 } {
	regexp {^[ ]*([-a-zA-Z0-9?!:,._]+)(.*)} $line match firstTag Rest
	if {[info exists formatTable($firstTag)]} {
	  if { $formatTable($firstTag) == "Tree" } {
	    # Tag: <firstTag_N> [<Rest>...</Rest>] </firstTag_N>
	    if {[regexp {[a-zA-Z0-9]+} $Rest]} {
	      StartTag $Rest
	    }
	    TagNow $firstTag
	    FindAllTags
	    killSegmentationWindow
	  } else {
	    # Tag: <firstTag_N="Rest">...</firsTag>
	    # or: <firstTag_N>...</firstTag_N>
	    TagNow "$firstTag $Rest"
	    FindAllTags
	    killSegmentationWindow
	  }
	} else {
	  TagNow "$firstTag $Rest" 
	  FindAllTags
	  killSegmentationWindow
	}
      } else {
	WarningMessage "StartTag: No text seclection?"
      }
    } else {
      WarningMessage "StartTag: No tag?"
    }
   
  }
  
}

set topMenu $this_menu.tag.menu
set topSequence ""
set doneList [list]
set tearoff true
foreach topTag $choiceTable(TopLevel) {
    if {[lsearch -exact $doneList $topTag] == -1} {
      AddMenu $topTag $topMenu $topSequence $font1 $bg1
      lappend doneList $topTag
    }
}


set this_menu $menuArea.edit_tag.menu
menu $this_menu -bg  $bg1  -tearoff false


$this_menu add command -label "Undo Last" -font $font1 \
    -command Untag 


set replaceTag 0


$this_menu add command -label "Change One..." -font $font1 \
    -command {
      wm title .chwindow "Nb  Edit Tags: Reading Text Tags..."
      wm deiconify .chwindow
      raise .chwindow .
      set editingAction  "Edit Tags"
      set editingMessage "Select A Tag"
      set renameMessage  "Change To..."
      set entryTag ""
      set oldTag   ""
      set replaceTag 0
      $bchoicesArea.add configure -state disabled -foreground gray
      $typeArea.value configure -state normal
      $typeArea.label configure -state normal
      $bchoicesArea.delete configure -state normal
      $bchoicesArea.rename configure -state normal
      if {[$textArea.text tag ranges ViewTags.sel] != ""} {
	$textArea.text tag delete ViewTags.sel
      }
      FindAllTags
      wm title .chwindow "Nb  Edit Tags"
      # update idletasks
    }


$this_menu add command -label "Delete All" -font $font1 \
    -command {
      while {[llength $clipBoard] > 0} {
        Untag
      }
    }

$this_menu add command -label "Email address..." -font $font1 \
    -command {
      EnterEmailAddress
    }

#$this_menu add command -label "Add Choice" -font $font1 \
#    -command {
#      wm title .chwindow "Nb  Add Choice"
#      wm deiconify .chwindow
#      raise .chwindow .
#      set entryTag ""
#      set oldTag   ""
#      set replaceTag 0
#      set editingMessage "Select a menu"
#      set renameMessage  "and type in a new choice"
#      $typeArea.value configure -state normal 
#      $typeArea.label configure -state disabled -disabledforeground $fg1
#      $bchoicesArea.delete configure -state disabled
#      $bchoicesArea.rename configure -state disabled
#      $bchoicesArea.add configure -state normal -foreground black
#      if {[$textArea.text tag ranges ViewTags.sel] != ""} {
#	$textArea.text tag delete ViewTags.sel
#      }
#      set editingAction "Add New Tag"
#      FindMenuWidgets
#      wm title .chwindow "Nb  Add Choice"
#       update idletasks
#    }


proc FindMenuWidgets {} {
    global editingAction menuWidgets choiceTable choicesArea
    set editingAction "Add New Tag"
    set collectList [lsort [array names menuWidgets]]
    $choicesArea.directory delete 0 end
    foreach thisId $collectList {
      $choicesArea.directory insert end $thisId
    }
}



set this_menu $menuArea.view.menu
$this_menu add command -label "Segment List" -font $font1 -command view_color_palette

source $NB_SOURCE/find_text.tcl
$this_menu add command -label "Find Text" -font $font1 -command CallFindText


# $this_menu add command -label "References   " -font $font1 \
#    -command {
#  # Highlight tags with the same identifier idTable(Tag)
#  puts "Click: View References"
#    }


set sizeBuffer 0
set textBuffer(0)    ""
set output_file_name ""

# View Output File used to be here. it's not here any more.
# You can view the output file with Emacs or another text editor...


proc FindSpecificTag { fromIndex toIndex listTags } {
    global textArea fieldTable choiceTable idTable beginAtIndex endAtIndex

    set index       [$textArea.text index "$fromIndex - 1 chars"]
    set toSel       [IndexToInt [$textArea.text index "$toIndex + 1 chars"]]
    set result      ""
    set spaces      " "
    set toIndex     [$textArea.text index "$toIndex"]
    while {$index != $toIndex} {
   if {[info exists beginAtIndex($index)]} {
       foreach thisTag $beginAtIndex($index) {
      if {[info exists idTable($thisTag)]} {
          regexp {^([-a-zA-Z0-9]+)\_([0-9]+)} $thisTag match head number
          if {[lsearch -exact $listTags $head] != -1} {
         set toTag [IndexToInt [lindex [$textArea.text tag ranges $thisTag] 1]]
         if {$toTag <= $toSel} {
             if {[regexp {^([-a-zA-Z0-9]+)\_([0-9]+)[ ]+([^ ].+)$} $idTable($thisTag) match head number name] != 0} {
            set result $result$spaces$name
             } else {
            if {[regexp {^([-a-zA-Z0-9]+\_[0-9]+)$} $idTable($thisTag) match value] != 0} {
                set result $result$spaces$value
            }
             }
         }
          }
      }
       }
   }
   set index [$textArea.text index "$index + 1 chars"]
    }
    return $result
}

set fileCommand ""

proc RunCommandOnFile { commandLine directory fileName } {
    global NB_SOURCE
    puts "Exec: $commandLine on dir: $directory and file: $fileName"
    if {![file executable $commandLine]} {
   set commandLine $NB_SOURCE/$commandLine
    }
    if {[file executable $commandLine]} {
   if {[file readable $directory$fileName]} {
       ### set unixOutput [exec $commandLine $directory$fileName &]
       WarningMessage "Option not implemented yet."
       .d configure -cursor arrow
       catch {wm withdraw .d}
   } else {
       WarningMessage "Can't read file: $fileName in directory: $directory"
   }
    } else {
   WarningMessage "$commandLine is not an executable command. Check permissions or path name in configuration file."
    }
}


# Always execute commandLine, possibly adding as arguments some specific tag names
# that are found in the annotated text selected by the user... 
proc CallCommand { commandTags commandLine } {
  global textArea fileFunction
  global NB_SOURCE current_directory choiceTable fileLabel fileValue fileCommand
  set argument ""
  set spaces " "
  
  if {![file executable $commandLine]} {
    set commandLine $NB_SOURCE/$commandLine
    if {![file executable $commandLine]} {
      WarningMessage "Can't execute $commandLine. Check permissions or path name in configuration file."
    }
  }
  if {[info exists choiceTable($commandTags)]} {
    if {[lindex $choiceTable($commandTags) 0] == "SelectFile"} {
      set fileLabel    "Find"
      set fileValue    ""
      set fileCommand  "$commandLine"
      set fileFunction "Open"
      wm deiconify .d 
      raise .d .
      DisplayFiles .d.m.directory $current_directory
      return
    }
  }
  if {[$textArea.text tag ranges sel] != ""} {
    if {[info exists choiceTable($commandTags)]} {
      foreach tagType $choiceTable($commandTags) {
	set one_argument [FindSpecificTag [$textArea.text index "sel.first"] [$textArea.text index "sel.last"] $tagType]
	if {[string length $one_argument] > 0} {
          set argument $argument$spaces$one_argument
	}
      }
      if {[string length $argument] > 0} {
	### set unixOutput [exec $commandLine $argument &]
      } else {
	### set unixOutput [exec $commandLine &]
      }
    } else {
      ### set unixOutput [exec $commandLine &]
    }
  } else {
    if {[string length $commandLine] > 0} {
      ### set unixOutput [exec $commandLine &]
    }
  }
}


# Add menus with external command specifications:
set numCommands 0
set suffix "Tags"


if {[info exists choiceTable(Commands)]} {
    foreach type $choiceTable(Commands) {
   regexp {^([-a-zA-Z0-9]+)Commands} $type match header
   set commandTags $header$suffix
   if {[info exists choiceTable($type)]} {
       foreach com $choiceTable($type) {
      if {[info exists formatTable($com)]} {
          incr numCommands
          puts "View Command: $com = $formatTable($com)"
          $this_menu add command -label $com -font $font1 -command [list CallCommand $commandTags $formatTable($com)]
      }
       }
   }
    }
}




set whiteTags [list]
if {[info exists choiceTable(WhiteTags)]} {
    foreach choice $choiceTable(WhiteTags) {
      if {[info exists choiceTable($choice)]} {
	set whiteTags [concat $whiteTags $choiceTable($choice)]
      }
      lappend whiteTags $choice
    }
}

set refTags [list]
set refCounter 0
set refTable(0) 0
set isRefTag(0) 0

set NbTags [list]
if {[info exists choiceTable(NbTags)]} {
    foreach choice $choiceTable(NbTags) {
        if {[info exists choiceTable($choice)]} {
            set NbTags [concat $whiteTags $choiceTable($choice)]
        }
        lappend NbTags $choice
    }
}
set whiteAndNbTags [concat $whiteTags $NbTags]

if {[info exists choiceTable(RefTags)]} {
    foreach choice $choiceTable(RefTags) {
      lappend refTags $choice
    }
}


puts "Reference Tags:   $refTags"
puts "Nb Tags:          $NbTags"
puts "Non visible Tags: $whiteTags"


# Help on the various buttons and menus:
set this_menu $menuArea.help.menu

MakeHelpMenu


$this_menu add command -label "About Nb " -font $font1 \
    -command {
   puts "Click: Help About Nb"
   set warningMessage \
       "  Nb - Nota bene 
  A visual discourse segmentation tool
  
  Copyright (c) 1995-1998 
  Massachusetts Institute of Technology
  Laboratory for Computer Science
  Spoken Language Systems Group
  Giovanni Flammia

  Send comments to: flammia@sls.lcs.mit.edu
  Up-to-date documentation and free downloads at:
  http://www.sls.lcs.mit.edu/flammia/Nb.html"

      wm deiconify .third
      raise .third

    }



# pack $menuArea.file $menuArea.tag $menuArea.view $menuArea.edit_tag $menuArea.mode $menuArea.optionValue
# -side left -expand 0
pack $menuArea.file $menuArea.tag $menuArea.view $menuArea.font $menuArea.edit_tag -side left -expand 0

pack $menuArea.help -side right -expand 0

# Secondary Pop-Up window with text for source file and other stuff:
toplevel  .second -class frame -bg $bg1 -relief ridge -colormap .dbox -visual $colorMap
wm  title .second "Nb  Second Window"
wm  iconname .second "Nb  Second Window"
frame .second.top -bg $bg1
frame .second.mid -bg $bg1
frame .second.bot -bg $bg1

pack .second.top -side top -expand 1 -fill both


text      .second.top.text    -relief sunken -font $text_font  -bg $text2_bg -fg $text2_fg \
    -yscrollcommand ".second.top.yscroll set" \
    -xscrollcommand ".second.mid.xscroll set" \
    -state normal \
    -height [expr $annotated_text_lines] -wrap none -setgrid 1 
scrollbar .second.top.yscroll -orient vertical  -command ".second.top.text yview" -relief sunken -bg $scroll_bg 
scrollbar .second.mid.xscroll -orient horizontal -command ".second.top.text xview" -relief sunken -bg $scroll_bg 

pack      .second.top.yscroll -side right -fill y 
pack      .second.top.text    -side left -fill both -expand 1 

set pad [expr [.second.top.yscroll cget -width] + 2* \
        ([.second.top.yscroll cget -bd] + \
        [.second.top.yscroll cget -highlightthickness])]
frame .second.mid.pad -width $pad -height $pad -bg $bg1

pack   .second.mid -side top -fill x
pack   .second.mid.pad -side right
pack   .second.mid.xscroll -side right -fill x -expand true

frame  .second.bot.pad -width $pad -height $pad -bg $bg1
button .second.bot.dismiss -text "Dismiss" -font $font1 -command "wm withdraw .second" -relief raised -bg $bg1 -fg $fg1 -padx 20


 
pack   .second.bot -side bottom -fill x
pack   .second.bot.pad -side right
pack   .second.bot.dismiss -side right 

# Third window is a Message Pop-up window for warning messages, copyright notices and so on...
source $NB_SOURCE/message_functions.tcl
toplevel  .third  -class frame -bg $bg1 -relief ridge -colormap .dbox -visual $colorMap

wm  title .third "Nb  Message"
wm  iconname .third "Nb  Message"


frame .third.top    -bg $bg1
frame .third.middle -bg $bg1
frame .third.bot    -bg $bg1

pack .third.top    -side top -expand 0 
pack .third.middle -side top -expand 1 -fill both
pack .third.bot    -side top -expand 0

## label .third.top.logo -bitmap @$NB_SOURCE/nb_logo.xbm -bg $bg1 -fg $fg1 
## pack  .third.top.logo -side top -expand 0

message .third.middle.msg -aspect 400 -justify left -padx 50 -pady 50 -bg $bg1 -fg $fg1 -font $font1 \
    -textvariable warningMessage 
pack .third.middle.msg -expand 1 -fill both

set warningMessage "Copyright (c) 1995 Massachusetts Institute of Technology"

button .third.bot.dismiss -text "Dismiss" -font $font1 -command "wm withdraw .third" -relief raised -bg $bg1 -fg $fg1 -padx 20
pack   .third.bot.dismiss -side right




# find all tags for currently annotated text and put them in cacheTag() cacheRange() tables
# cacheTag(IdSting) returns the Tag identified by idString
# cacheRange(Tag) returns a list of two elements with the range of text tagged with Tag
proc FindAllTags {} {
    global textArea cacheRange cacheTag 
    global whiteAndNbTags idTable choicesArea currentTags currentIndex
    $choicesArea.directory delete 0 end

    foreach thisTag $currentTags {
        if {[info exists idTable($thisTag)]} {
            regexp {^([a-zA-Z0-9]+)} $thisTag match head
            if {[lsearch -exact $whiteAndNbTags $head] == -1} {
                $choicesArea.directory insert end $idTable($thisTag)
                set cacheRange($thisTag) [$textArea.text tag ranges $thisTag]
                set cacheTag($idTable($thisTag)) $thisTag
                set cacheTag($thisTag) $thisTag
            }
        }
    }
    $choicesArea.directory insert end " "
}


# Call editing window if tag is under the cursor:
proc EditTagsUnderCursor {} {

    global found_tags 
    global textArea cacheRange cacheTag 
    global whiteAndNbTags typeArea tagUnderCursor idUnderCursor idTable choicesArea currentTags currentIndex
    global editingAction editingMessage renameMessage entryTag oldTag bchoicesArea
    global selectedId oldTag fileExists entryTag replaceTag cursor_position
    

    $choicesArea.directory delete 0 end
    set editingAction  "Edit Tags"
    set editingMessage "Select Tag"
    set renameMessage  "Change To..."
    set entryTag       ""
    set oldTag         ""
    set replaceTag     0
    set lastTag        "notag"

    $bchoicesArea.add configure -state disabled -foreground gray
    $typeArea.value configure -state normal
    $typeArea.label configure -state normal
    $bchoicesArea.delete configure -state normal
    $bchoicesArea.rename configure -state normal
    
    set found 0
    set listTags [split $idUnderCursor \n]
    
    foreach thisId $listTags {
      if {[regexp {([a-zA-Z0-9]+)} $thisId match ]} {
        regexp {([-_a-zA-Z0-9]+\_[0-9]+)} $thisId match thisTag
        if {[string compare $thisTag $lastTag] != 0} {
	  if {[info exists idTable($thisTag)]} {
	    regexp {^([a-zA-Z0-9]+)} $thisTag match head
	    set lastTag $thisTag
	    if {[lsearch -exact $whiteAndNbTags $head] == -1} {
	      $choicesArea.directory insert end $idTable($thisTag)
	      set cacheRange($thisTag) [$textArea.text tag ranges $thisTag]
	      set cacheTag($idTable($thisTag)) $thisTag
	      set cacheTag($thisTag) $thisTag
	      if {$found==0} {
		set firstTag $thisTag
	      }
	      set found [expr $found + 1]
	    }
	  }
        }
      }
    }
    
    $choicesArea.directory insert end " "

    if {$found > 0} {
      set replaceTag 1
      set found_tags $found
      if {$found == 1} {
	# this one is not working under MacOS unfortunately...
        $choicesArea.directory activate 0
        $choicesArea.directory selection set 0 0
        set selectedId $idTable($firstTag)
        set oldTag     $firstTag
        set fileExists 1
        set thisTag    $firstTag
        set entryTag   ""
        if {[regexp {^([-a-zA-Z0-9]+)\_([0-9]+)[ ]+([^ ].+)$} $selectedId match head number identifier] == 1} {
           set entryTag $identifier
        } else {
           set entryTag ""
        }
        set fromIndex [lindex $cacheRange($firstTag) 0]
        set toIndex   [lindex $cacheRange($firstTag) 1]
        $textArea.text tag add ViewTags.sel $fromIndex $toIndex
        $textArea.text tag configure ViewTags.sel -background LightSeaGreen -foreground yellow

      } 
    } else {
      set found_tags 0
      catch {wm withdraw .chwindow}
    }

    update idletasks
    puts "Passed: EditTagsUnderCursor..."
    
  }


set selectedId ""
set oldTag     ""
set currentTag ""
set lastIndex  0
set touchedIndex 0

# pack $menuArea $typeArea $choicesArea -side top
# bind <Double-Button-1> to edit the current Tag
bind $choicesArea.directory <ButtonRelease-1> {
  if {[$textArea.text tag ranges ViewTags.sel] != ""} {
    $textArea.text tag delete ViewTags.sel
  }
  set value      ""
  set selectedId ""
  set oldTag     ""
  set entryTag   ""
  set indices    [list]
  if [catch {set indices [$choicesArea.directory curselection]}] {
    set indices [list]
  }
  if {[llength $indices] > 0} {
    set touchedIndex [lindex $indices 0]
    set value [lindex [$choicesArea.directory get [lindex $indices 0]] 0]
    set oldTag $value
    set entryTag ""
    if {[lindex $indices 0] < 5} {
      set lastIndex 0
    } else {
      set lastIndex [expr [lindex $indices 0] - 4]
    }
  } else {
    set value      ""
    set selectedId ""
    set oldTag     ""
    set entryTag   ""
  }
  if {$value != ""} {
    if {$editingAction == "Edit Tags"} {
      if {[info exists idTable($value)]} {
	set selectedId  $idTable($value)
	set oldTag      $value
	set fileExists  1
	set thisTag     $value
	set entryTag    ""
	set replaceTag  1
	set fromIndex   [lindex $cacheRange($thisTag) 0]
	set toIndex     [lindex $cacheRange($thisTag) 1]
	$textArea.text  tag add ViewTags.sel $fromIndex $toIndex
	$textArea.text  tag configure ViewTags.sel -background LightSeaGreen -foreground yellow
	$textArea.text  see $fromIndex
	if {[regexp {^([-a-zA-Z0-9]+)\_([0-9]+)[ ]+([^ ].+)$} $selectedId match head number identifier] == 1} {
	  set entryTag $identifier
	} else {
	  set entryTag ""
	}
      }
    } else {
      if {$editingAction == "Add New Tag"} {
	set replaceTag 0
	if {$oldTag != ""} {
	  if {[info exists menuWidgets($oldTag)]} {
	    WarningMessage "Selected Menu: \"$oldTag\""
	  }
	}
      }
    }
  }
}


# Rename tag with a brand new identifier or add a new tag to some pull-down menu:
bind $typeArea.value <Return> { 
    if {$editingAction == "Edit Tags"} {
      .chwindow config -cursor watch
      set editingAction "Name Tag"
      EditMenuHandler 
      set editingAction "Edit Tags"
      FindAllTags
      set entryTag ""
      set oldTag   ""
      .chwindow config -cursor arrow
    } else {
      EditMenuHandler
      set entryTag ""
      set editingAction "Add New Tag"
    }
  if {[$textArea.text tag ranges sel] != ""} {
    # Do nothing for the moment
  } else {
    if {[$textArea.text tag ranges ViewTags.sel] != ""} {
      $textArea.text tag delete ViewTags.sel
   }
  }
}








# Selection handler: actions to take when the user highlight some text:
# Now everything is handled by TagNow in edit_functions.tcl

set defaultWordTag "NP"
set defaultLineTag "Segment"
set wordTagIndex 0
set lineTagIndex 0
set underscore "_"
set fileLabel ""
set fileValue ""
set fileCommand ""

# File Selection Pop-Up window is a listbox called ".d":
# create a listbox:
toplevel  .d -class frame -bg $bg1 -relief ridge -colormap .dbox -visual $colorMap
wm  title .d "Nb  File Window"
wm  iconname .d "Nb  File Window"


# label and entry: Directory: [$current_directory_path]
frame .d.l -bg $bg1
frame .d.f -bg $bg1
frame .d.m -bg $bg1
frame .d.b -bg $bg1 
label .d.l.dir_label -text "Directory" -anchor sw -width 12 -font $font1 -bg $bg1 -fg $fg1 
entry .d.l.dir_value -textvariable current_directory -selectforeground $text_bg -selectbackground $text_fg \
    -font $font1 -relief sunken -bg $listbox_bg -fg $listbox_fg -width 55
.d.l.dir_value config  -insertbackground yellow

label .d.f.file_label -textvariable fileLabel -anchor sw -width 12 -font $font1 -bg $bg1 -fg $fg1
entry .d.f.file_value -textvariable fileValue \
    -font $font1 -relief sunken -bg $listbox_bg -fg $listbox_fg -width 55
.d.f.file_value config  -insertbackground yellow

# .d.l.dir_value xview [expr [.d.l.dir_value index end] - 50]

bind .d.f.file_value <Return> \
    { 
      switch $fileLabel {
        "Find"
        { 
          FindFileOrPattern .d.m.directory $fileValue
          catch {wm withdraw .dbox}
        }
        "Save File"
        {    
          .d configure -cursor watch
          update idletasks
          SaveFile
          wm title . "Nb [file tail $output_file_name]"
          .d configure -cursor arrow
          catch {wm withdraw .d}
          catch {wm withdraw .dbox}
          
        }
        
      }
    }


   
bind .d.l.dir_value <Return> \
    {
   puts "current_directory is $current_directory" 
   if {[file isdirectory $current_directory] == 1} {
      
      # set system_output [exec /bin/cd $current_directory]  
      .d configure -cursor watch
      catch {wm withdraw .dbox}
      update idletasks
      DisplayFiles .d.m.directory $current_directory
      .d configure -cursor arrow
      update idletasks
       
   } else {
       WarningMessage "I can't find a directory called: $current_directory"
   }
   # .d.l.dir_value view [expr [.d.l.dir_value index end] - 50] 
    }
pack .d.l .d.f -side top -expand 1 -fill x 
pack .d.m -side top -expand 1 -fill both
pack .d.b -side bottom -expand 1 -fill x
pack .d.l.dir_label -side left
pack .d.l.dir_value -side left -fill x -expand 1
pack .d.f.file_label -side left
pack .d.f.file_value -side left -fill x -expand 1

# listbox with the current_directory listing:
listbox   .d.m.directory -relief raised -yscrollcommand ".d.m.scroll_directory set" \
    -bg $listbox_bg -fg $listbox_fg -font $font1 -width 65 -height 8 \
    -exportselection 0 \
    -selectbackground white -selectforeground DarkSlateBlue -highlightcolor white
scrollbar .d.m.scroll_directory -command ".d.m.directory yview" -bg $scroll_bg -relief sunken
pack      .d.m.scroll_directory -side right -fill y 
pack      .d.m.directory -side left -fill both -expand 1 

set fileFunction "Open" 

button .d.b.doit -textvariable fileFunction -font $font1 -command "DoFileFunction" -relief raised -bg $bg1 -fg $fg1 -padx 20
button .d.b.dismiss -text "Dismiss" -font $font1 -command "wm withdraw .d" -relief raised -bg $bg1 -fg $fg1 -padx 20
pack   .d.b.dismiss -side right
pack   .d.b.doit -side left

# bind <Double-Button-1> and .d.b.doit to either {DisplayFiles selection} or to {LoadFile selection}:
# or to {RunCommandOnFile fileCommand selection}:
proc DoFileFunction {} {
    
    global tagUnderCursor fileFunction current_directory input_file_name output_file_name
    global fileLabel fileValue user_selected rangeTable beginAtIndex endAtIndex
    global listOfSgmlTags fileExists colorTable fileCommand textArea sgmlFormat
    
    set indices [list]
    set user_selected ""
    if [catch {set indices [.d.m.directory curselection]}] {
        set indices [list]
    }
    if {[llength $indices] > 0} {
        set user_selected [lindex [.d.m.directory get [lindex $indices 0]] 0]
    }
    set slash "/"
    if { $fileLabel == "Find" } {
      if { $user_selected != "" } {
        # If user selects a directory, list that directory in the listbox:
        set choicedir 0
        if  { [string index $user_selected [expr [string length $user_selected] - 1]] == "/" } {
          set choicedir 1
        } 
        if  { [string compare $user_selected ".."] == 0 } {
          set choicedir 2
        }
        if { $choicedir > 0 } {
          if { $choicedir == 2 } {
            set current_directory [my_file_dirname $current_directory]
          } else {
            if { [string index $current_directory [expr [string length $current_directory] - 1]] == "/" } {
              set current_directory "$current_directory$user_selected"
            } else {
              set current_directory "$current_directory/$user_selected"
            }
          }
          DisplayFiles .d.m.directory $current_directory
        } else {
          if { [string index $current_directory [expr [string length $current_directory] - 1]] != "/" } {
            set current_directory $current_directory$slash
          }  
          # ThenLoadFile fills rangeTable:
          regsub {[@\*]$} $user_selected "" clean_selection
          puts "Dir: $current_directory File: $clean_selection"
          if {[string length $fileCommand] > 0} {
            RunCommandOnFile $fileCommand $current_directory $clean_selection
          } else {
            # else try to call LoadFile with argument $user_selected:
            .d configure -cursor watch
            update idletasks
            set tagUnderCursor "Loading: [file tail $clean_selection]"
            ThenLoadFile $current_directory$clean_selection
            .d configure -cursor arrow
          }
        }
      } else {
        if {$fileValue != ""} {
          FindFileOrPattern .d.m.directory $fileValue
          return
        }
      }
    } else {
        if { $fileLabel == "Save File" } {
          SaveFile 
          catch {wm withdraw .d}
        }
    }
  }


bind .d.m.directory <Double-Button-1> {
  if {[string compare $fileLabel "Save File"]!=0} {
    DoFileFunction
  }
}




# Initial state: withdraw all PopUp windows:

# return value of colormode:
wm withdraw .d
wm withdraw .chwindow
wm withdraw .second
wm withdraw .third 
wm withdraw .dbox
wm withdraw .view
wm withdraw .outline
wm withdraw .colors
wm withdraw .id

set refCounter 0
set refTable(0) 0

# Load InputFile?
if { $argc >= 4 } {
  puts "Current Directory is: ( $current_directory )"
  puts "Requested File is: ( [lindex $argv 3] )"
  ThenLoadFile [lindex $argv 3] 
}



proc EnterEmailAddress {} {

  global fileLabel fileValue fileCommand fileFunction output_file_name 
  global input_file_name annotatorName suffix configPrefix current_directory
  global dboxVariable dboxEntry dboxMessage dboxOk dboxDismiss
  
  set dboxOk        "Ok"
  set dboxDismiss   "Quit Program"
  set dboxVariable  $current_directory
  set dboxMessage   "Enter your email address for getting the gift certificates and click Ok"
  set fileLabel     "Email Address"
  set fileValue     ""
  set fileCommand   ""
  set fileFunction  "Email"
  
  wm deiconify  .dbox
  set dboxEntry "$annotatorName"
  wm deiconify  .dbox
  raise .dbox   .
  update idletasks

}

### Uncomment this line if you want that the
### user sends the annotated files to flammia@sls.lcs.mit.edu
### To change the settings of email server/user to send to,
### edit the source file send_mail.tcl
### EnterEmailAddress
