#####################################################################################################
# FILE:    bindings.tcl
# PURPOSE: Customize keyboard bindings for easier editing of entries and text widgets
# AUTHOR:  Giovanni Flammia
#
#          (N)[b] Copyright (c) 1995
#          Spoken Language Systems Group
#          Laboratory for Computer Science
#          Massachusetts Institute of Technology
#####################################################################################################



set bindingsBuffer ""


# returns "Success" if success, else return error message.
proc send_mail {files toaddress mailserver1 mailserver2} {
  
  global annotatorName tagUnderCursor 
  
  set success 0  
  set mailserver $mailserver1

  # open client socket connection to sendmail server on port 23
  if [catch {set s [socket $mailserver1 25]}] {
    puts "Failure to open socket connection to mail server: $mailserver1"
    set mailserver $mailserver2
    if [catch {set s [socket $mailserver2 25]}] {
      puts "Failure to open socket connection to mail server: $mailserver2"
      return "Failure to open socket connection to mail server: $mailserver2"
    } 
  }
  
  fconfigure $s -buffering line -translation {auto lf}
  if {![eof $s]} {
    
    puts "Connected to mail server: $mailserver..."
    set tagUnderCursor "Connected to mail server..."
    update idletasks

    set senderhost "nb-annotator.lcs.mit.edu"
    if [catch {set senderhost [info hostname]}] {
      set senderhost "nb-annotator.lcs.mit.edu"
    }
  
    set sendername [pid]
    set sendername "$annotatorName"
    if {![regexp {[a-zA-Z]+} $sendername match]} {
      set sendername [pid]
    }

    bell
    update idletasks
    puts "Sending mail header..."
    
    puts "Setting my host to: $senderhost"
    puts "Setting my user name to: $sendername"
    puts "Setting to address to: $toaddress"

    puts $s "HELO $senderhost"
    puts $s "MAIL FROM: $sendername"
    puts $s "RCPT TO: $toaddress"
    puts $s "DATA"
    puts $s "Subject: $senderhost $sendername"
    flush $s
    
    foreach filename $files {
      if {[file readable $filename]} {
        if {![eof $s]} {
          set file [open $filename r]
          puts "Sending file: $filename..."
          puts $s "<File=$filename User=$sendername Host=$senderhost>"
          while {[gets $file line] >= 0} {
            if {[regexp {[^ ]+} $line match]} {
              puts $s "$line"
              #### puts "$line"
              flush $s
            }
            if {[eof $s]} {
              puts "Connection closed prematurely by mail server: $mailserver."
              return "Connection closed prematurely by mail server: $mailserver."
            }
          }
          close $file
          bell
          update idletasks
        } else {
          puts "Connection closed prematurely by mail server: $mailserver."
          return "Connection closed prematurely by mail server: $mailserver."
        }
      } else {
        
        puts "Sending mail footer prematurely..."
        puts $s "."
        puts $s "QUIT"
        flush $s
        close $s
        
        puts "Cannot find annotated file: $filename."
        return "Cannot find annotated file: $filename."
        
      }
    }
    
    if {![eof $s]} {
      
      puts "Sending mail footer..."
      puts $s "."
      puts $s "QUIT"
      flush $s
      close $s
      
      set success 1
      puts "Done!"
      set tagUnderCursor "Done!"
      update idletasks

    } else {
      
      puts "Connection closed prematurely by mail server: $mailserver."
       return "Connection closed prematurely by mail server: $mailserver."
      
    }
  }
  
  if {$success == 1} {
    puts "Success!"
    return "Success!"
  } else {
    puts "Failure!"
    return "Failure!"
  }

}



proc MoveLeft {w} {
  set result [expr [$w index insert] - 1]
  if {$result <= 0} {
    return 0
  }
  return $result
}


proc DeleteLeft {w} {
  set result [expr [$w index insert] - 1]
  if {$result <= 0} {
    $w delete 0
  }
  $w delete $result
}


proc MoveRight {w} {
  set result [expr [$w index insert] + 1]
  if {$result >= [$w index end]} {
    return [$w index end]
  }
  return $result
}


bind Entry <KeyPress-Delete> { DeleteLeft %W ; break }
bind Entry <KeyPress-Left>   { %W icursor [MoveLeft %W] ; break }
bind Entry <KeyPress-Right>  { %W icursor [MoveRight %W] ; break }
bind Text  <KeyPress-u>      { Untag }
bind Text  <KeyPress-Delete> { Untag }
bind Text  <KeyPress-BackSpace> { Untag }
bind Text  <KeyPress-v>      { view_color_palette }

proc TextMoveLeft {t} {
    $t mark set insert "insert - 1 chars"
}


proc TextDeleteLeft {t} {
  $t delete "insert - 1 chars"
}


proc TextMoveRight {t} {
  $t mark set insert "insert + 1 chars"
}



proc DeleteRestLine {t} {
  global bindingsBuffer
  set bindingsBuffer [$t get "insert" "insert lineend"]
  $t delete "insert" "insert lineend"
}


proc YankSelection {t} {
    global bindingsBuffer
    if {$bindingsBuffer != ""} {
   $t insert "insert" $bindingsBuffer
    } else {
   WarningMessage "Yank: text selection is empty. Use <Control-w> or <Control-k> to select some text to yank."
    }
}


proc WipeSelection {t} {
    global bindingsBuffer
    if {[$t tag ranges sel] != ""} {
   set bindingsBuffer [$t get "sel.first" "sel.last"]
   $t delete "sel.first" "sel.last"
    }
}


bind Text <KeyPress-Delete> { TextDeleteLeft %W ; break }
bind Text <KeyPress-Left> { TextMoveLeft %W ; break }
bind Text <KeyPress-Right> { TextMoveRight %W ; break }
bind Text <Control-k> { DeleteRestLine %W ; break }
bind Text <Control-y> { YankSelection %W ; break }
bind Text <Control-w> { WipeSelection %W ; break }


# Special custom Tag bindings from config file:
# Example configuration file:
# 
#FIELD 
#  Bindings
#CHOICES
#  <KeyPress-Tab>     Segment
#  <ButtonPress-3>    Segment
#       <KeyPress-n>       NP
#       <KyePress-v>       VP
#HELP
#  Custom bindings for annotating segments and phrases
#END

proc CustomBind {} {
  global choiceTable formatTable textArea
  if {[info exists choiceTable(ShortcutsForTags)]} {
    foreach binding $choiceTable(ShortcutsForTags) {
      if {[string compare $binding "<Motion>"] != 0} {
        if {[string compare $binding "<ButtonPress-2>"] != 0} {
          if {[info exists formatTable($binding)]} {
          if [catch {bind $textArea.text $binding "StartTag \"$formatTable($binding)\" "}] {
            puts "Invalid binding, did not bind: $binding to anything!"
          }
          }
        } else {
          WarningMessage "Cannot bind <ButtonPress-2> in Configuration file. It is already bound to edit tags. Sorry."
        }
      } else {
        WarningMessage "Cannot bind <Motion> in Configuration file. It is alreay bound to show tags under cursor. Sorry."
      }
    }
  }
}



proc CustomColorBind {} {
  
  global choiceTable formatTable colorForId colorTable otherColorForId
  
  if {[info exists choiceTable(ColorsForTags)]} {
    foreach binding $choiceTable(ColorsForTags) {
      if {[info exists formatTable($binding)]} {
        if {[regexp {^([-\_a-zA-Z0-9]+)[ ]+(.+)$} $formatTable($binding) match head name]} {
          set colorForId($head.$name) $binding

        } else {
          set colorTable($formatTable($binding)) $binding
        }
      }
    }
  }
  
  # Alternate colors for tags that are next to each other...
  if {[info exists choiceTable(OtherColorsForTags)]} {
    foreach binding $choiceTable(OtherColorsForTags) {
      if {[info exists formatTable($binding)]} {
        if {[regexp {^([-\_a-zA-Z0-9]+)[ ]+(.+)$} $formatTable($binding) match head name]} {
          set otherColorForId($head.$name) $binding

        } else {
          set colorTable($formatTable($binding)) $binding
        }
      }
    }
  }
}



proc ReBind {} {
    CustomBind
    CustomColorBind
}
 
proc hyperlink_handler {arg} {
    
    global textArea hyper textBuffer refCounter tagUnderCursor annotatorName

    selection clear $textArea.text 
    catch {$textArea.text tag delete ViewTags.sel}
    catch {$textArea.text tag delete ViewTags.copy}
    update idletasks

    puts "Called hyperlink handler: $arg"
    if [catch {regexp -- {^([a-zA-Z0-9]+)[ ]+([-._~//\\a-zA-Z0-9]+)} $arg match action value}] {
        puts "Hyperlink handler: cannot parse argument $arg"
    } else {   
       if [catch {puts "Hyperlink handler: action $action, value: $value"}] {
          WarningMessage "No action for: $arg" 
       } else {
       switch -regexp -- $action {
         ^Quit 
         {
           QuitProgram
         }
         ^Send
         {

           set flag 0
           FirstSaveFile
           set tagUnderCursor "Attempting to send annotated files..."
           set annotated_files [list]
           for {set i 1} {$i <= 10} {incr i} {
             if {[file readable "Ex-$i"]} {
               lappend annotated_files "Ex-$i"
             }
           }
           for {set i 1} {$i <= 120} {incr i} {
             if {[file readable "Annotation-$i"]} {
               lappend annotated_files "Annotation-$i"
             }
           }
           WarningMessage "Attempting to send files: $annotated_files by email, please wait..."
           bell
           update idletasks
           # set mailserver1 "18.27.0.171"
           set mail1 "18.69.0.28" 
           set mail2 "18.72.1.2"
           set toaddress   "flammia@sls.lcs.mit.edu"
           if [catch {set success [send_mail $annotated_files $toaddress $mail1 $mail2]}] {
             set flag 1
           }
           bell
           update idletasks
           if {$flag == 0} {
             if {[regexp {Success} $success match]} {
               WarningMessage "I sent your annotated files by email to flammia@sls.lcs.mit.edu"
             } else {
               WarningMessage "$success"
             }
           } else {
             WarningMessage "I was unable to send your annotated files by email. Sorry."
           }
         }
	 ^Message
	 {
	   set message "";
	   if {[file exists $value]} {
	     set file [open $value r] 
	     while {[gets $file line] >= 0} {
	       set message "$message\n$line"
	     }
	     close $file
	     WarningMessage $message
	   } else {
	     WarningMessage "File: $value not found."
	   }
	 }
	 ^LoadHelp
	 {
	   if {[file readable $value]} {
	     catch {unset textBuffer}
	     set textBuffer(0) ""
	     LoadTextFile   $value
	     ShowSecondText .second.top.text
	     wm deiconify   .second
	     wm title       .second "Nb Help"
	     wm iconname    .second "Nb Help"
	   } else {
	     WarningMessage "File: $value not found."
	   }
	 }
	 ^Load 
	 {
	   set hyper 1
	   FirstSaveFile    
	   ThenLoadFile  $value
	   $textArea.text see 1.0
	   selection clear $textArea.text 
	   catch {$textArea.text tag delete ViewTags.sel}
	   catch {$textArea.text tag delete ViewTags.copy}
           update idletasks
           if {[string compare $value "Ex-1"] == 0} {
             if [catch {exec ./nb.bat &}] {
               puts "Failed to spawn a new Nb window with the command: ./nb.bat"
             }
           }
	 }
	 ^Open 
	 {
	   
	   set hyper 1
	   FirstSaveFile    
	   ThenLoadFile  $value
	   $textArea.text see 1.0
	   update idletasks
	   selection clear $textArea.text 
	   catch {$textArea.text tag delete ViewTags.sel}
	   catch {$textArea.text tag delete ViewTags.copy}
	 }
	 ^Save 
	 {   

	   set hyper 1
	   FirstSaveFile
	   selection clear $textArea.text
	   update idletasks 
	   catch {$textArea.text tag delete ViewTags.sel}
	   catch {$textArea.text tag delete ViewTags.copy}
	 }
         ^View
         {
           set hyper 1
           view_color_palette
           update idletasks
         }
	 default 
	 {
	   puts "Hyperlink handler: no action accociated with $action"
	   selection clear $textArea.text
	   update idletasks
	  
	 }
       }
       }
    }
  }


puts "Loading: bindings.tcl..."
CustomBind
CustomColorBind
