###################################################################
# FILE:    edit_functions.tcl
# PURPOSE: Tag and text editing functions for Nb
# AUTHOR:  Giovanni Flammia 
#
#         (N)[b] Copyright (c) 1995
#         Spoken Language Systems Group
#         Laboratory for Computer Science
#         Massachusetts Institute of Technology
###################################################################
puts "Loading: edit_functions.tcl..."

# produce a tagged text from $fromIndex to $toIndex:
proc TaggedText {fromIndex toIndex} {
  global whiteTags textArea idTable dmTags choiceTable beginAtIndex endAtIndex
  set from [IndexToInt $fromIndex]
  set to   [IndexToInt $toIndex]
  set spaces " "
  puts "Non Parse Tags: $whiteTags"
  if { $from < $to } {
    set currentIndex $fromIndex
    set outputString ""
    set bracketCount 0
    set openTags [list]
    set endSelection [$textArea.text index "$toIndex"]
    while {$currentIndex != $endSelection} {
      # Prefix <Tag="identifier(Tag)"> ?
      if {[info exists beginAtIndex($currentIndex)]} {
        if {[llength $beginAtIndex($currentIndex)] > 0} {
          foreach oneTag $beginAtIndex($currentIndex) {
            if {$idTable($oneTag) == $oneTag} {
              set concatString "<$oneTag>"
            } else {
              regexp {^([-a-zA-Z0-9]+)\_([0-9]+)[ ]+([^ ].*)$} $idTable($oneTag) match head number identifier
              set concatString  "<$oneTag=\"$identifier\">"
            }
            set outputString $outputString$concatString
          }
        }
      }
      # write down char indexed:
      set concatString [$textArea.text get $currentIndex]
      regsub -all {[{}<>]} $concatString "_" subString
      set concatString $subString
      if { $concatString != "" } {
        if { $concatString != "\n" } {
          set outputString $outputString$concatString
        } else {
          set outputString $outputString$spaces
        }
      } else {
        set outputString $outputString$spaces
      }
      set currentIndex [$textArea.text index "$currentIndex + 1 chars"]
      # Suffix </Tag> ?
      if {[info exists endAtIndex($currentIndex)]} {
        if {[llength $endAtIndex($currentIndex)] > 0} {
          # Print the tags in reverse order to respect indentation:
          set t [expr [llength $endAtIndex($currentIndex)] - 1]
          while {$t >= 0} {
            set oneTag [lindex $endAtIndex($currentIndex) $t]
            set concatString "</$oneTag>"
            set outputString $outputString$concatString
            set t [expr $t - 1]
          }
        }
      }
    }
  } else {
    return "Selection does not show tagged text"
  }
  return "\"$outputString\""
}



# Select a color for the new tag that is not the same with neighboring or
# overlapping tags. Segment tags are assigned all different colors, while
# phrase and word tags are assigned a color dependent on the headTag 
# That is: all NP have one color, VP have another color and so on...
# However, if a tag has an Tag Name identifier like "NP_1 identifier" then the color
# is based on taghead.identifier (for coloring co-references with the same color)
# Return: color of $tagName + list of other tags within fromIndex and toIndex
# Note: idName is "tagName identifier" or "tagName" 
proc SelectColor { tagName idName fromIndex toIndex } {
  
  global currentTags text_fg default_color 
  global tagUnderCursor cacheFrom cacheTo rangeTable 
  global segmentColor colorIndex idTable colorForId otherColorForId 
  global colorTable textArea fieldTable choiceTable
  global beginAtIndex endAtIndex segments

  set realFromIndex   [IndexToInt [$textArea.text index "$fromIndex - 1 chars"]]
  set forbiddenColors [list]
  set listOfOtherTags [list]
  set nearbyTags      [list]

  ### $textArea.text see $fromIndex
  update idletasks

  set thisEnd    [IndexToInt [$textArea.text index "$toIndex"]]
  set thisBegin  [IndexToInt [$textArea.text index "$fromIndex"]]
  set pointer    [lsearch -exact $currentTags $tagName]
  set last       [expr [llength $currentTags] - 1]

  set firstPointer $pointer
  if {$firstPointer < 0} {
    set firstPointer 0
  }

  set lastPointer [expr $pointer + 2]
  if {$lastPointer > $last} {
    set lastPointer $last
  }
  
  regexp {^([-a-zA-Z0-9]+)\_([0-9]+)[ ]+(.+)$} $idName match this_head this_number this_name
  regexp {^([0-9]+)\.} [$textArea.text index "$fromIndex"] match first_row
  regexp {^([0-9]+)\.} [$textArea.text index "$toIndex"] match last_row
  
  for {set i $firstPointer} {[expr $i <= $lastPointer]} {incr i} {
    set otherTag [lindex $currentTags $i]
    if {[expr [string length $otherTag] > 1]} {
      if {[expr [expr [expr $cacheFrom($otherTag) >= $realFromIndex] && [expr $cacheTo($otherTag) < $thisEnd]] || [expr [expr $cacheFrom($otherTag) > $realFromIndex] && [expr $cacheTo($otherTag) <= $thisEnd]]] } {
        lappend listOfOtherTags $otherTag
        if {[info exists colorTable($otherTag)]} {
          lappend forbiddenColors $colorTable($otherTag)
        }
      }
      if {[expr $cacheTo($otherTag) > $thisEnd]} {
        break
      }
    }
    if {[expr $i > $last]} {
      break
    }
  }
  
  # Check whether this tag follows another tag with the same color:
  set tagHead   $tagName
  set name      ""
  set number    ""
  set head      ""
  set alreadyExists 0
  
  regexp {^([-a-zA-Z0-9]+)\_([0-9]+)$} $tagName match tagHead value
  regexp {^([-a-zA-Z0-9]+)\_([0-9]+)[ ]+(.+)$} $idName match head number name
  
  set other_head   ""
  set other_number ""
  set other_name   ""
  set change_color "no"
  set indexBefore  [expr $first_row - 1]
  # hack, can be improved...
  set indexBefore  [$textArea.text index "$indexBefore.0 lineend -4 chars"]
  set indexAfter   [expr $last_row + 1]
  set indexAfter   [$textArea.text index "$indexAfter.0 lineend -4 chars"]
  set nearbyTags   [concat [$textArea.text tag names $fromIndex] [$textArea.text tag names $toIndex]]
  set nearbyTags   [concat $nearbyTags [$textArea.text tag names $indexBefore]]
  set nearbyTags   [concat $nearbyTags [$textArea.text tag names $indexAfter]]
  
  #puts "$tagName       : $fromIndex $toIndex"
  #puts "nearby_tags    : $indexBefore $indexAfter"
  #puts "nearby tags are: ($nearbyTags)"

  if {![string match "." "$head.$name"]} {
    if {[info exists otherColorForId($head.$name)]} {
      if {[info exists choiceTable(SegmentTags)]} {
        if {[lsearch -exact $choiceTable(SegmentTags) $head] != -1} {
          foreach oneTag $nearbyTags {
            if {[info exists idTable($oneTag)]} {
              #puts "Checking nearby tag: $idTable($oneTag)"
              regexp {^([-a-zA-Z0-9]+)\_([0-9]+)[ ]+(.+)$} $idTable($oneTag) match other_head other_number other_name
              if {[string compare "$head.$name" "$other_head.$other_name"] == 0} {
                if {[string compare $idTable($tagName) $idTable($oneTag)] != 0} {
                  set xfromIndex [lindex $rangeTable($oneTag) 0]
                  set xtoIndex   [lindex $rangeTable($oneTag) 1]
                  regexp {^([0-9]+)\.([0-9]+)} $xfromIndex match fromRow fromCol
                  regexp {^([0-9]+)\.([0-9]+)} $xtoIndex match toRow toCol              
                  set change_color $colorTable($oneTag)
                  #puts "$tagName must change color from $colorTable($oneTag) because of $oneTag"
                  break
                } else {
                  #puts "$idTable($oneTag) is the same as $idTable($tagName)"
                }
              } else {
                #puts "$idTable($oneTag) is different from $idTable($tagName)"
              }
            }
          }
        }
      }
      if {[string compare $change_color $colorForId($head.$name)] == 0} {
        set temp $colorForId($head.$name)
        set colorForId($head.$name)      $otherColorForId($head.$name)
        set otherColorForId($head.$name) $temp
      } 
    }
  }
  
  
  if {[info exists colorTable($tagHead)]} {
    set alreadyExists 1
  }
  if {[info exists choiceTable(SegmentTags)]} {
    if {[lsearch -exact $choiceTable(SegmentTags) $tagHead] != -1} {
      set alreadyExists 0
    }
  }
  if {[string length $name] > 1} {
    if {[info exists colorForId($head.$name)]} {
      set colorTable($tagName) $colorForId($head.$name)
      set segmentColor($tagName) $colorTable($tagName)
      return [concat [list $colorTable($tagName)] $listOfOtherTags]
    } else {
        if {$alreadyExists == 1} {
          set colorTable($tagName)   $colorTable($tagHead)
          set segmentColor($tagName) $colorTable($tagHead)
          return [concat [list $colorTable($tagName)] $listOfOtherTags]
        }  
    }
    set alreadyExists 0
  }
  if {[info exists choiceTable(WhiteTags)]} {
    if {[lsearch -exact $choiceTable(WhiteTags) $tagHead] != -1} {
      ### puts "WhiteTags ( $WhiteTags ) contains $tagHead, so put color to $default_color"
      set colorTable($tagHead) $default_color
      set alreadyExists 1
    }
  }
  if {$alreadyExists == 1} {
    set colorTable($tagName) $colorTable($tagHead)
    set segmentColor($tagName) $colorTable($tagName)
    return [concat [list $colorTable($tagName)] $listOfOtherTags]
  }
  if {[info exists choiceTable(Colors)]} {
    set  savedIndex $colorIndex
    incr colorIndex
    if {$colorIndex >= [llength $choiceTable(Colors)]} {
      set colorIndex 0
    }
    while {[lsearch -exact $forbiddenColors [lindex $choiceTable(Colors) $colorIndex]] != -1} {
      incr colorIndex 
      if {$colorIndex == $savedIndex} {
        break
      }
      if {$colorIndex >= [llength $choiceTable(Colors)]} {
        set colorIndex 0
      }
    }
  } else {
    WarningMessage "SelectColor: You have to define the palette of available Colors in the configuration file."
    return [concat [list 1] $listOfOtherTags]
  }
  set colorTable($tagName) [lindex $choiceTable(Colors) $colorIndex]
  ### set colorTable($tagHead) $colorTable($tagName)
  if {[string length $name] > 1} {
    set colorForId($head.$name) $colorTable($tagName)
  }
  set segmentColor($tagName) $colorTable($tagName)
  return [concat [list $colorTable($tagName)] $listOfOtherTags]
}


proc CheckForRefTags {ref id from to} {
  
  global beginAtIndex endAtIndex cacheFrom CacheTo idTable
  global refTags refTable refCounter tagUnderCursor statusMessage
  
  set key "$from $to"

  if {[info exists refTable($key)]} {  
    set m             ""
    set refPattern    ""
    set testPattern   ""
    set oldRefCounter $refCounter
    regexp {^Ref([a-zA-Z]+)[_ 0-9]+([a-zA-Z0-9].+)$} $refTable($key) m refType refName
    regexp {^([a-zA-Z]+)[_ 0-9]+([a-zA-Z0-9].+)$} $id m testType testName
    if {[string compare $testType $refType] == 0} {
      if {[string compare $testName $refName] == 0} {
        set statusMessage "Bravo! Found $refType."
        return 1
      } else {
        set statusMessage "Boundaries of a $refType with another name ($refCounter left)."
        return 0
      }
    } else {
      set statusMessage "Boundaries of a $refType ($refCounter left)."
      bell
      return 0
    }
  }
  if {$refCounter > 0} {
    set statusMessage "No matches ($refCounter left)"
  }
  return 0

}


# Tag current selection with new tag whose id will be topLevelTag
# TagNow grabs the text widget selection and does not clear it
#
# If: $topLevelTag = "FirstTag RestTags"
# Then: <FirstTag_Number="RestTags">fromIndex...toIndex</FirstTag_Number>
#
#            id string  
proc TagNow { topLevelTag } {

  global cacheFrom cacheTo currentTags currentIndex option clipBoard
  global insertIndex tagUnderCursor textArea text_fg text_bg refTags refCounter refTable
  global whiteTags choicesArea fileChanged fileExists rangeTable
  global formatTable beginAtIndex choiceTable endAtIndex textSelection rowSelection
  global replaceTag selectedId cacheTag cacheRange NbTags
  global editingAction fieldTable colorTable idTable statusMessage
  global allowManyTags isRefTag
 

  set underscore   "_"
  set replacedTag  ""
  set oldIndex     -1
  set tagUnderCursor "Tagging: $topLevelTag"
  set idUnderCursor  "$topLevelTag"
  #puts "Entering: TagNow $topLevelTag replace this tag: $replaceTag"
  if {$replaceTag == 1} {
     if {$selectedId != ""} {
         if {[info exists cacheTag($selectedId)]} {
	   set replacedTag $cacheTag($selectedId)
	   regexp {(^[-_a-zA-Z0-9]+)\_([0-9]+)} $replacedTag match oldHead oldIndex
	   if {![regexp ^$oldHead $topLevelTag]} {
	     set oldIndex -1
	   }
	   set fromIndex   [lindex $cacheRange($replacedTag) 0]
	   set toIndex     [lindex $cacheRange($replacedTag) 1]
	   if {[$textArea.text tag ranges ViewTags.copy] == ""} {
	     $textArea.text tag add ViewTags.copy $fromIndex $toIndex
	   }
	   DeleteThisTag
	   set replaceTag 0
         }
      }
  }
  update idletasks
  if {[$textArea.text tag ranges ViewTags.copy] != ""} {
    # Clean-up selection:
    set fromIndex [lindex [$textArea.text tag ranges ViewTags.copy] 0]
    set toIndex   [lindex [$textArea.text tag ranges ViewTags.copy] 1]
    set spaces "    "
    regexp {^([0-9]+)\.([0-9]+)} $fromIndex match rowStart colStart
    set rowSelection $rowStart
    regexp {^([0-9]+)\.([0-9]+)} $toIndex match rowEnd colEnd
    set eol 0
    # Check for true end-of-line: 
    while {[expr [expr $rowStart < $rowEnd] && [expr $colEnd <= 1]] == 1} {
      set toIndex [$textArea.text index "$toIndex - 1 chars"]
      regexp {^([0-9]+)\.([0-9]+)} $toIndex match rowEnd colEnd
      # puts "toIndex: $toIndex row: $rowEnd col: $colEnd"
      set eol 1
    } 
    while {[expr [expr [string match " " [$textArea.text get $fromIndex]] || [string match "\n" [$textArea.text get $fromIndex]]] && ![info exists beginAtIndex($fromIndex)]]} {
      set fromIndex [$textArea.text index "$fromIndex + 1 chars"]
    }
    while {[expr [expr [string match " " [$textArea.text get [$textArea.text index "$toIndex - 1 chars"]]] || [string match "\n" [$textArea.text get [$textArea.text index "$toIndex - 1 chars"]]]] && ![info exists endAtIndex($toIndex)]]} {
      set toIndex [$textArea.text index "$toIndex - 1 chars"]
    }
    switch $option {
      "Tag Selection"
      {
         # Tag exactly the highlighted area and nothing else:
      } 
      "Tag Words" 
      {
         # Tag all the non-space words highlighetd by the selection:
         set fromIndex [$textArea.text index "$fromIndex wordstart"] 
         if {$eol == 0} {
            set toIndex   [$textArea.text index "$toIndex wordend"]
         }
      }  
      "Tag Lines" 
      {
         # Tag all the lines highlighted by the selection:
         set fromIndex [$textArea.text index "$fromIndex linestart"]
         if {$eol == 0} {
            set toIndex   [$textArea.text index "$toIndex lineend"]
         }
      }
      default 
      {
         WarningMessage "Tag: You have to set the option to: Tag Selection, Tag Words or Tag Lines"
         return ""
      }
    }
    while {[expr [expr [string match " " [$textArea.text get $fromIndex]] || [string match "\n" [$textArea.text get $fromIndex]]] && ![info exists beginAtIndex($fromIndex)]]} {
      set fromIndex [$textArea.text index "$fromIndex + 1 chars"]
    }
    while {[expr [expr [string match " " [$textArea.text get [$textArea.text index "$toIndex - 1 chars"]]] || [string match "\n" [$textArea.text get [$textArea.text index "$toIndex - 1 chars"]]]] && ![info exists endAtIndex($toIndex)]]} {
      set toIndex [$textArea.text index "$toIndex - 1 chars"]
    }
    regexp {^([0-9]+)\.([0-9]+)} $toIndex match rowEnd colEnd
    set eol 0
    # Check for true end-of-line: 
    while {[expr [expr $rowStart < $rowEnd] && [expr $colEnd <= 1]] == 1} {
      set toIndex [$textArea.text index "$toIndex - 1 chars"]
      regexp {^([0-9]+)\.([0-9]+)} $toIndex match rowEnd colEnd
      # puts "toIndex: $toIndex row: $rowEnd col: $colEnd"
      set eol 1
    } 
    set textSelection ""
    if {$allowManyTags == 0} {
      if {[info exists beginAtIndex($fromIndex)]} {
	if {[info exists endAtIndex($toIndex)]} {
	  foreach other_tag $beginAtIndex($fromIndex) {
	    if {![info exists isRefTag($other_tag)]} {
	      if {[lsearch -exact $endAtIndex($toIndex) $other_tag] != -1} {
		WarningMessage "This text is already tagged with: $idTable($other_tag). Click the mouse right
  button to change or delete a Segment or SubSegment tag."
		return ""
	      }
	    }
	  }
	}
      }
    }
    #puts "passed TagNow $topLevelTag $fromIndex $toIndex"
    if {[expr [string compare $fromIndex $toIndex] != 0]} {
      set crossBrackets [CrossingBrackets $topLevelTag $fromIndex $toIndex]
      if {$crossBrackets != ""} {
        if {[regexp {^Bracketing} $crossBrackets]} {
          WarningMessage "$crossBrackets"
          bell
          return ""
        }
	if {[regexp {^Ref} $crossBrackets match]} { 
	  set crossBrackets " some reference tag"
	}
	WarningMessage "Your text selection crosses the end-points of $crossBrackets"
	bell
	return ""
      } else {
	regexp {^([-0-9a-zA-Z\!\?\_\-\:\.\;\,]+)} $topLevelTag firstTag
	regsub {([ ]+)$} $topLevelTag "" completeTag
	if {$refCounter > 0} {
	 
	  if {[lsearch -exact $refTags "Ref$firstTag"] != -1} {
	    set match [CheckForRefTags "Ref$firstTag" $completeTag $fromIndex $toIndex]
	    if {$match == 0} {
	      puts "reference tags not found"
	      return ""
	    } else {
	      set refCounter [expr $refCounter - 1]
	      if {$refCounter > 0} {
		set statusMessage "Bravo! $refCounter tags left."
		bell
	      } else {
		set statusMessage "All Done!"
		bell
		bell
	      }
	    }
	  }
	} 
	set fileChanged 1
	set fileExists  1
	for {set c $fromIndex} {$c != $toIndex} {set c [$textArea.text index "$c + 1 chars"]} {
	  set textSelection "$textSelection[$textArea.text get $c]"
	}
	regsub -all {([][$\"\?;:,!\{\}+\.])} $textSelection "" textSelection
	# puts "text selection: $textSelection"
	# Extract first keyword:
	# Lookup current used index
	if {![info exists fieldTable($firstTag)]} {
	  set fieldTable($firstTag) [list 0]
	}
	if {$oldIndex > 0} {
	  set xindex $oldIndex
	  set lastIndex [lindex $fieldTable($firstTag) 0] 
	  if { $lastIndex <= $oldIndex } {
	    set fieldTable($firstTag) [list $oldIndex]
	  }    
	} else {
	  set xindex 0
	  set xindex [lindex $fieldTable($firstTag) 0]
	  incr xindex
	  set  fieldTable($firstTag) [list $xindex]
	}
	# Change tag:
	set newTag $firstTag$underscore$xindex
	# set new color for this tag:
	set colorTable($newTag) "white"
	# in $topLevelTag replace $firstTag with $newTag and put in idString
	regsub ^$firstTag $completeTag $newTag idString
	# puts "Selection: ($completeTag) Tag: ($newTag) Id: ($idString)"
	set idTable($newTag) $idString
	set cacheFrom($newTag) [IndexToInt $fromIndex]
	set cacheTo($newTag) [IndexToInt $toIndex]
	set currentIndex($newTag) $insertIndex
	$textArea.text tag add $newTag $fromIndex $toIndex
	set rangeTable($newTag) [list $fromIndex $toIndex]
	set currentTags [InsertInPriorityList $newTag $currentTags]
	set colorList [SelectColor $newTag $idString $fromIndex $toIndex]
	# Insert tag end-points in priority lists:
	set plist [list]
	if {[info exists beginAtIndex($fromIndex)]} {
	  set plist $beginAtIndex($fromIndex)
	}
	set beginAtIndex($fromIndex) [InsertInPriorityList $newTag $plist]
	set plist [list]
	if {[info exists endAtIndex($toIndex)]} {
	  set plist $endAtIndex($toIndex)
	}
	if {[llength $clipBoard] > 0} {
	  set clipBoard [linsert $clipBoard 0 $newTag]
	} else {
	  set clipBoard [list $newTag]
	}
	set endAtIndex($toIndex) [InsertInPriorityList $newTag $plist]
	# puts "Begin List \($beginAtIndex($fromIndex)\)"
	# puts "End   List \($endAtIndex($toIndex)\)"
	# [lindex $colorList 0] is the color assigned for this tag:
	$textArea.text tag configure $newTag -foreground [lindex $colorList 0] 
	if {[ lsearch -exact $NbTags $firstTag ] != -1} {
	  if {[regexp {^[-_.a-zA-Z0-9]+[ ]+([a-zA-Z].+)} $idTable($newTag) match rest]} {
	    #puts "Binding tag: $newTag link to [lindex $colorList 0]" 
	    $textArea.text tag bind $newTag <ButtonPress-1> "hyperlink_handler \"$rest\" ; break "
	    $textArea.text tag configure $newTag -underline true
	    $textArea.text tag bind $newTag <Enter> "$textArea.text configure -cursor arrow"
	    $textArea.text tag bind $newTag <Leave> "$textArea.text configure -cursor xterm"
	  } 
	}    
        update idletasks
	# extract list of other tags:
	set colorList  [lreplace $colorList 0 0]
	set thisBegin  [IndexToInt [lindex [$textArea.text tag ranges $newTag] 0]]
	set thisEnd    [IndexToInt [lindex [$textArea.text tag ranges $newTag] 1]]
	# check whether there is a refTags to match.
	set noRaise [list]
        set underscore "_"
	if {[info exists choiceTable(WhiteTags)]} {
            set noRaise $choiceTable(WhiteTags)
	}
	# raise binding priority of other tags:
	foreach otherTag $colorList {
	  if {[regexp {^([-a-zA-Z0-9]+)\_([0-9]+)$} $otherTag match head value] >= 1} {
	    if {[string compare $otherTag $newTag] != 0} {
              #puts "Raising other tag: $otherTag"
	      if {[lsearch -exact $noRaise $head] == -1} {
		set otherBegin [IndexToInt [lindex [$textArea.text tag ranges $otherTag] 0]]
		set otherEnd   [IndexToInt [lindex [$textArea.text tag ranges $otherTag] 1]]
		if {[expr [expr $otherBegin >= $thisBegin] && [expr $otherEnd <= $thisEnd]]} {
                  set s "$head$underscore$value"
		  $textArea.text tag raise $otherTag
                  $textArea.text tag raise hidden.$s
                  $textArea.text tag raise hidden.color.$s
		}
	      }
	    }
	  }
	}
        update idletasks
	if {[info exists choiceTable(SegmentTags)]} {
	  if {[lsearch -exact $choiceTable(SegmentTags) $firstTag] != -1} {
	    AddSegment $newTag $fromIndex $toIndex online
            puts "Added segment: $newTag"
	  }
	}
        update idletasks
	return $newTag
      }
    }
  } else {
    WarningMessage "Tag: You have to select some text first"
    return ""
  }
}

  


   

# List all the current idTable(Tags) under the cursor:
proc ListTagsUnderCursor {} {
  global idTable tagUnderCursor idUnderCursor textArea whiteTags
  
  set listId  ""
  set listTag ""
  set spaces " "
  set newline "
"
  set dots "..."
  set l [$textArea.text tag names "current"]
  set n [expr [llength $l] - 1]
  for {set i $n} {$i >= 0} {set i [expr $i - 1]} {
    set oneTag [lindex $l $i]
    if {[info exists idTable($oneTag)]} {
      regexp {^([a-zA-Z0-9]+)\_} $oneTag match head
      if {[lsearch -exact $whiteTags $head] == -1} {
        if {[string compare $head "Highlight"] != 0} {
          set listId  $listId$idTable($oneTag)$newline
          if {[expr [string length $idTable($oneTag)] < 80]} {
            set word "$idTable($oneTag) "
          } else {
            set word "$oneTag "
          }
          if {[expr [string length $listTag] < 80]} {
            set listTag $listTag$word
          } else {
            set listTag $listTag$dots
          }
        }
      }
    }
  }
  # this one is used for display
  regsub -all {[_ ]+} $listTag " " tagUnderCursor
  # this one is used for editing data strucures
  regsub -all {[ ]+}  $listId  " " idUnderCursor
}


# returns the tag that crosses brackets between {fromIndex toIndex}
# else returns the empty string
proc CrossingBrackets {tagName fromIndex toIndex} {
  global textArea idTable maxLevelOfNesting
  set from [IndexToInt $fromIndex]
  set to   [IndexToInt $toIndex]
  set nearbyTags [$textArea.text tag names $fromIndex]
  set nearbySize [llength $nearbyTags]
  set line "$nearbyTags"
  if {[regexp {RefS} $line match]} {
    return ""
  }
  while {[regexp {hidden.(.+)} $line match rest]} {
    set nearbySize [expr $nearbySize - 1]
    set line $rest
  }
  if {[regexp {sel} $nearbyTags match]} {
    set nearbySize [expr $nearbySize - 1]
  }
  if {[regexp {ViewTags} $nearbyTags match]} {
    set nearbySize [expr $nearbySize - 1]
  }
  if {[regexp {SubSegment} $tagName match]} { 
    if {![regexp {Segment} $nearbyTags match]} {
      return "Bracketing problem: a SubSegment must be contained in a Segment."
    } 
    if {$nearbySize > $maxLevelOfNesting} {
      return "Bracketing problem: too many nested SubSegments (maximum should be $maxLevelOfNesting)"
    }
  } else {
    if {[regexp {Segment} $tagName match]} {
      if {[regexp {Segment} $nearbyTags match]} {
        return "Bracketing problem: a Segment cannot be nested in another Segment (use SubSegment instead)."
      }
      if {[regexp {SubSegment} $nearbyTags match]} {
        return "Bracketing problem: a top-level Segment cannot be nested in a SubSegment. "
      }
    }
  }
  regexp {^([0-9]+)} $fromIndex match thisrow
  foreach thisTag [concat $nearbyTags [$textArea.text tag names $toIndex]] {
    if {[info exists idTable($thisTag)]} {
      regexp {^([0-9]+)} [lindex [$textArea.text tag ranges $thisTag] 0] match thatrow
      if {$thisrow == $thatrow} {
        return "Bracketing problem: the starting line already starts $idTable($thisTag). A line can either start a Segment or a SubSegment but not both."
      }
      set begin [IndexToInt [lindex [$textArea.text tag ranges $thisTag] 0]]
      set end   [IndexToInt [lindex [$textArea.text tag ranges $thisTag] 1]]
      if {[expr $begin < $from] && [expr $end < $to]} {
        puts "$begin smaller than $from $fromIndex and $end smaller than $to $toIndex\n";
        if {[expr $from - $begin] < [expr $to - $end]} {
          return "$idTable($thisTag)"
        } else {
          return "$idTable($thisTag)"
        }
      }
      if {[expr $from < $begin] && [expr $to < $end]} {
        puts "$begin bigger than $from $fromIndex and $end bigger than $to $toIndex\n";
        if {[expr $begin - $from] < [expr $end - $to]} {
          return "$idTable($thisTag)"
        } else {
          return "$idTable($thisTag)"
        }
      }
    }
  }
  return ""
}


# Convert a text widget index (row.col) into an positive integer value
# Assume a max number of chars per line of 512
proc IndexToInt {textIndex} {
  if { $textIndex != "" } {
    if {[regexp {^([0-9]+)\.([0-9]+)$} $textIndex match rowIndex colIndex]} {
      return [expr $rowIndex*512 + $colIndex]
    } else {
      return -1
    }
  } else {
    return -1
  }
}


# Return >0 if tag1 should come before tag2
#        <0 if tag2 should come before tag1
#        0  if tag1 and tag2 have the same end-points
# Assuming tags do not cross brackets
proc PreOrder { tag1 tag2 } {
  global textArea
  set from1 [IndexToInt [$textArea.text index [lindex [$textArea.text tag ranges $tag1] 0]]]
  set to1   [IndexToInt [$textArea.text index [lindex [$textArea.text tag ranges $tag1] 1]]]
  set from2 [IndexToInt [$textArea.text index [lindex [$textArea.text tag ranges $tag2] 0]]]
  set to2   [IndexToInt [$textArea.text index [lindex [$textArea.text tag ranges $tag2] 1]]]
  if {$from1 == $from2} {
    return [expr $to1 - $to2]
  }
  return [expr $from2 - $from1]
}


proc FastPreOrder { tag1 tag2 } {
  global textArea cacheFrom cacheTo
  if {[info exists cacheFrom($tag1)]} {
    set from1 $cacheFrom($tag1)
  } else {
    set from1 [IndexToInt [$textArea.text index [lindex [$textArea.text tag ranges $tag1] 0]]]
    set cacheFrom($tag1) $from1
  }
  if {[info exists cacheTo($tag1)]} {
    set to1 $cacheTo($tag1)
  } else {
    set to1   [IndexToInt [$textArea.text index [lindex [$textArea.text tag ranges $tag1] 1]]]
    set cacheTo($tag1) $to1
  }
  if {[info exists cacheFrom($tag2)]} {
    set from2 $cacheFrom($tag2)
  } else {
    set from2 [IndexToInt [$textArea.text index [lindex [$textArea.text tag ranges $tag2] 0]]]
    set cacheFrom($tag2) $from2
  }
  if {[info exists cacheTo($tag2)]} {
    set to2 $cacheTo($tag2)
  } else {
    set to2   [IndexToInt [$textArea.text index [lindex [$textArea.text tag ranges $tag2] 1]]]
    set cacheTo($tag2) $to2
  } 
  if {$from1 == $from2} {
    return [expr $to1 - $to2]
  }
  return [expr $from2 - $from1]
}

set insertIndex -1

# Insert tag in a priorityList respecting the preorder sequence:
# Returns the updated list:
# Average worst case running time: O(List length = Average number of tag end-points per text index)
proc InsertInPriorityList {tag pList} {
  global insertIndex
  set size [llength $pList]
  set i 0
  set iPoint 0
  while {$i < $size} {
    if {[FastPreOrder [lindex $pList $i] $tag] < 0} {
      set i $size
    } else {
      incr i
      set iPoint $i
    }
  }
  set insertIndex $iPoint
  return [linsert $pList $iPoint $tag]
}
   


# delete tag from priority list:
proc DeleteFromPriorityList {tag priorityList} {
  set deletePoint [lsearch -exact $priorityList $tag]
  if {$deletePoint != -1} {
    return [lreplace $priorityList $deletePoint $deletePoint]
  } else {
    return $priorityList
  }
}


# When a new sgml/Nb annotated file is loaded, the tables idTable colorTable and rangeTable
# should be filled with appropriate tag information. This is when you can call this
# procedure that adds all the tags listed in listOfTags, and fills the beginAtIndex
# and endAtIndex lists.
proc AddTagsFromTables { listOfTags } {

  global cacheFrom cacheTo refTags refCounter refTable
  global beginAtIndex whiteTags tagUnderCursor text_fg text_bg currentTags currentIndex
  global choiceTable refTable endAtIndex idTable colorTable rangeTable textArea fieldTable
  global NbTags isRefTag statusMessage segments 
  
  set currentTags [list]
  # Invisible tags: WhiteTags
  if {[info exists choiceTable(WhiteTags)]} {
    set noRaise $choiceTable(WhiteTags)
  }
  set tagUnderCursor "Updating Tags..."
  update idletasks
  set refCounter 0
  set noRaise [list]
  catch {unset refTable}
  catch {unset isRefTag}
  set refTable(0) 0
  foreach newTag $listOfTags {
    if {[info exists rangeTable($newTag)]} {
      if {[llength $rangeTable($newTag)] == 2} {
	set fromIndex [lindex $rangeTable($newTag) 0]
	set toIndex   [lindex $rangeTable($newTag) 1]
	set begin     [IndexToInt $fromIndex]
	set end       [IndexToInt $toIndex]
	if {$end >= $begin} {
	  set match ""
	  set value ""
	  set name ""
	  set crossBrackets ""
	  if {[regexp {^([-a-zA-Z0-9]+)_([0-9]+)} $newTag match name value]} {
	    if {[info exists fieldTable($name)]} {
	      set oldValue [lindex $fieldTable($name) 0]
	      if {$oldValue != ""} {
		if {$value > $oldValue} {
		  set fieldTable($name) [list $value]
		}
	      } else {
		set fieldTable($name) [list $value]
	      }
	    } else {
	      set fieldTable($name) [list $value]
	    }
	  } else {
	    WarningMessage "Warning: Tag <$newTag> does not have an Nb format!"
	    set fieldTable($name) [list 0]
	  }
	  lappend currentTags $newTag
	  if {[lsearch -exact $refTags $name] != -1} {
	    incr refCounter
	    set key "$fromIndex $toIndex"
	    set refTable($key) $idTable($newTag)
	    set isRefTag($newTag) 1
	  }
	  set currentIndex($newTag) [expr [llength $currentTags] - 1]
	  $textArea.text tag add $newTag $fromIndex $toIndex
	  if {[lsearch -exact $NbTags $name] != -1} {
	    if {[regexp {^[-_.a-zA-Z0-9]+[ ]+([a-zA-Z].+)} $idTable($newTag) match rest]} { 
	      $textArea.text tag bind $newTag <ButtonPress-1> "hyperlink_handler \"$rest\" ; break "
	      $textArea.text tag configure $newTag -underline true
	      $textArea.text tag bind $newTag <Enter> "$textArea.text configure -cursor arrow"
	      $textArea.text tag bind $newTag <Leave> "$textArea.text configure -cursor xterm"
	    } 
	  }    
	  set plist [list]
	  set cacheFrom($newTag) [IndexToInt $fromIndex]
	  set cacheTo($newTag)   [IndexToInt $toIndex]
	  if {[info exists beginAtIndex($fromIndex)]} {
	    set plist $beginAtIndex($fromIndex)
	  }
	  set beginAtIndex($fromIndex) [InsertInPriorityList $newTag $plist]
	  set plist [list]
	  if {[info exists endAtIndex($toIndex)]} {
	    set plist $endAtIndex($toIndex)
	  }
	  set endAtIndex($toIndex) [InsertInPriorityList $newTag $plist]
	  set colorList [SelectColor $newTag $idTable($newTag) $fromIndex $toIndex]
	  if {[lsearch -exact $noRaise $name] == -1} {
	    if {![info exists isRefTag($newTag)]} {
	      $textArea.text tag configure $newTag -foreground [lindex $colorList 0]
              #puts "Setting color for: $newTag to be [lindex $colorList 0]"
	    }
            update idletasks
	    $textArea.text tag raise $newTag
	    set colorList  [lreplace $colorList 0 0]
	    set thisBegin  [IndexToInt [lindex [$textArea.text tag ranges $newTag] 0]]
	    set thisEnd    [IndexToInt [lindex [$textArea.text tag ranges $newTag] 1]]
	    foreach otherTag $colorList {
	      if {[string compare $otherTag $newTag] != 0} {
                #puts "Found other tags that must be raised: ($otherTag)"
		if {[regexp {^([-a-zA-Z0-9]+)\_([0-9]+)$} $otherTag match head value] >= 1} {
		  if {[lsearch -exact $noRaise $head] == -1} {
                    set otherBegin [IndexToInt [lindex [$textArea.text tag ranges $otherTag] 0]]
                    set otherEnd [IndexToInt [lindex [$textArea.text tag ranges $otherTag] 1]]
                    if {[expr [expr $otherBegin >= $thisBegin] && [expr $otherEnd <= $thisEnd]]} {
		      $textArea.text tag raise $otherTag
                    }
		  }
		}
	      }
	    }
            if {[info exists choiceTable(SegmentTags)]} {
              if {[lsearch -exact $choiceTable(SegmentTags) $name] != -1} {
                AddSegment $newTag $fromIndex $toIndex batch
                puts "Added segment $newTag"
              }
            }
            update idletasks
	  }
	} else {
	  WarningMessage "Ignoring <$newTag>: Tag end-points are incorrect!"
	}
      } else {
	WarningMessage "Ignoring <$newTag>: Incorrectly closed or Incorrectly open!"
      }
    }
  }

  foreach newTag $currentTags {
    if {$refCounter > 0} {
      if {[regexp {^([-a-zA-Z0-9]+)_([0-9]+)} $newTag match name value]} {
	if {[lsearch -exact $refTags "Ref$name"] != -1} {
	  set match [CheckForRefTags "Ref$name" $idTable($newTag) [lindex $rangeTable($newTag) 0] [lindex $rangeTable($newTag) 1] ]
	  if {$match != 0} {
	    set refCounter [expr $refCounter - 1]
	  }
	}
      }
    } 
  }
  
  selection clear $textArea.text
  
  if {[$textArea.text tag ranges ViewTags.sel] != ""} {
    $textArea.text tag delete ViewTags.sel
  }
  
  if {$refCounter > 0} {
    set stausMessage "There are $refCounter hidden tags to find."
    WarningMessage "There are $refCounter hidden tags to find. When you find one, I will beep, and a message will appear in the bottom right corner of the Nb window. If you try to tag where there is no hidden tag, nothing will happen. "
  }
  
  set tagUnderCursor "Ready."
  update idletasks
  
  ##### bell
  
}

  

