#!/usr/local/bin/wish40

#####################################################################
# FILE:    show_segmentation.tcl
# PURPOSE: Display a colored segmentation in a text widget from input tagged file
#          This is the tcl/Tk program that is called when the user clicks on
#          "View Segments" in Nb, or "View Other"
# AUTHOR:  Giovanni Flammia 
#
# NOTE: If you dont specify otherwise in the command line arguments, this program
#       assumes that the segment tags are called <Segment_Number>...</Segment_Number>
#       (for example, when called from "View Other")
#
#         (N)[b] Copyright (c) 1995
#         Spoken Language Systems Group
#         Laboratory for Computer Science
#         Massachusetts Institute of Technology
#####################################################################


# Process input arguments:
if { $argc < 1 } {
    puts "Usage: show_segmentation.tcl TaggedFile \{Tag1 Color1 Tag2 Color2 ... \}
   Or: show_segmentation.tcl TaggedFile segmentTag
   Or: show_segmentation.tcl TaggedFile"
    exit
}

set colorTable(TopLevel) white
set inputFile            [lindex $argv 0]
set segmentTags          [list]

if {[llength $argv] > 1} {
    set segmentTags      [lindex $argv 1]
    set segmentType      ""
    for {set t 0} {$t < [llength $segmentTags]} {set t [expr $t + 2]} {
	if {[expr $t + 1] < [llength $segmentTags]} {
	    set colorTable([lindex $segmentTags $t]) [lindex $segmentTags [expr $t + 1]]
	}
    }
}

set segmentType "Segment"
if {[llength $segmentTags] > 1} {
    regexp {^([-_a-zA-Z0-9]+)\_[0-9]+$} [lindex $segmentTags 0] segmentType 
} else {
    if {[llength $segmentTags] == 1} {
	set segmentType [lindex $segmentTags 1]
    } else {
	set segmentType "Segment"
    }
}


# Name the window and global variables:
wm title . "(N)\[b\] View Segments"
wm iconname . "(N)\[b\] View Segments"


set bg1                "Grey"
set fg1                "Black"
set text_bg            "Black"
set text_fg            "White"
set text2_bg           "Grey"
set text2_fg           "Black"
set scroll_bg          "Black"
set listbox_bg         "Grey"
set listbox_fg         "Black"
set annotated_text_lines 30

toplevel .outline -class frame -bg $bg1 -relief raised
wm title .outline "(N)\[b\] Segmentation Outline"
wm iconname .outline  "(N)\[b\] Segmentation Outline"


# set font1              "lucidasanstypewriter-bold-14"
if {[info exists env(NB_FONT)]} {
    set font1 $env(NB_FONT)
} else {
    set font1              "-adobe-helvetica-bold-r-normal--18-180-75-75-p-103-iso8859-1"
}


# set font1                "-adobe-helvetica-bold-r-normal--24-240-75-75-p-138-iso8859-1"
if {[info exists env(NB_SOURCE)]} {
    set NB_SOURCE          $env(NB_SOURCE)
} else {
    puts "Can't find environment variable for NB_SOURCE" 
    exit
}


# Color palette for X terminals:
set palette [list honeydew3 \
		 plum \
		 sandybrown \
		 cyan \
		 LightSeaGreen \
		 Orange \
		 YellowGreen \
		 DeepPink \
		 PaleTurquoise \
		 khaki \
		 azure4 \
		 SeaGreen4 \
		 khaki4 \
		 LightYellow4 \
		 salmon3 \
		 aquamarine ]
set tagUnderCursor ""


proc WarningMessage { message } {
    puts $message
}

# Source files:
source $NB_SOURCE/file_functions.tcl
source $NB_SOURCE/output_functions.tcl

# Hold text and vertical scrollbar
frame .top -bg $bg1 

# Hold horizontal scrollbar and padx
frame .mid -bg $bg1   

# Hold dismiss and other buttons
frame .bot -bg $bg1

# Begin global variables:

# Hashtables for cascade-menu Tags:
# fieldTable(firstLevelTag) = list with currently used index for firstLevelTag
set fieldTable(TopLevel)  [list]

# choiceTable(Tag) = list with pull-down menu options for Tag
set choiceTable(TopLevel) [list]

# priority table for nested tags like: <TAG1> <TAG2> ... </TAG2> <TAG1>
set priorityTable(TopLevel) 0

# For each index, keep a list of tags in preorder:
set beginAtIndex(0.0) [list]
set endAtIndex(0.0) [list]

# Each tag has the format topLevelTag_number and is associated with
# a name in idTable (a string), a color in colorTable, and a range.
# Hastables for text Tags:
set listOfSgmlTags [list]
set idTable(TopLevel)     "TopLevel"

# Recently used color in fieldTable(Colors)
set colorIndex 0

# set rangeTable(TopLevel)  { "1.1" "1.1" }
set rangeTable(TopLevel) [list]

# End global variables

# Main window text widget: by default you can't write on it:
set textArea .top
text      $textArea.text   -relief sunken -font $font1 -bg $text_bg -fg $text_fg \
    -yscrollcommand "$textArea.yscroll set" \
    -xscrollcommand ".mid.xscroll set" \
    -exportselection 1 \
    -state normal \
    -height $annotated_text_lines -wrap none -setgrid 1 
scrollbar $textArea.yscroll  -orient vertical -command "$textArea.text yview" -relief raised -bg $scroll_bg
scrollbar .mid.xscroll  -orient horizontal -command "$textArea.text xview" -relief raised -bg $scroll_bg

frame .outline.1 -bg $bg1
frame .outline.2 -bg $bg1
frame .outline.3 -bg $bg1

set outlineLabel "Outline Label"

text .outline.1.text -relief sunken -font $font1 -bg $text_bg -fg $text_fg \
    -exportselection 1 \
    -yscrollcommand ".outline.1.yscroll set" \
    -xscrollcommand ".outline.2.xscroll set" \
    -state normal \
    -height $annotated_text_lines -wrap none -setgrid 1 
scrollbar .outline.1.yscroll  -orient vertical -command ".outline.1.text yview" -relief raised -bg $scroll_bg
scrollbar .outline.2.xscroll  -orient horizontal -command ".outline.1.text xview" -relief raised -bg $scroll_bg

button .outline.3.dismiss -text "Dismiss" -font $font1 -command "wm withdraw .outline" -relief raised -bg $bg1 -fg $fg1 -padx 20 
# label .outline.3.label -textvariable outlineLabel -font $font1 -fg Black -bg $bg1 -padx 20
    


# create padding based on size of scrollbars (p. 193 of Welch's book)
set pad [expr [$textArea.yscroll cget -width] + 2* \
	     ([$textArea.yscroll cget -bd] + \
		  [$textArea.yscroll cget -highlightthickness])]
frame .mid.pad -width $pad -height $pad -bg $bg1
frame .outline.2.pad -width $pad -height $pad -bg $bg1
frame .outline.3.pad -width $pad -height $pad -bg $bg1

button .bot.dismiss -text "Dismiss" -font $font1 -command exit -relief raised -bg $bg1 -fg $fg1 -padx 20 
frame .bot.pad -width $pad -height $pad -bg $bg1
label .bot.label -textvariable tagUnderCursor -font $font1 -fg Black -bg $bg1 -padx 20


# Arrange everything:


pack .bot -side bottom -fill x 
pack .bot.label -side left
pack .bot.pad -side right
pack .bot.dismiss -side right
pack .mid -side bottom -fill x
pack .mid.pad -side right
pack .mid.xscroll -side bottom -fill x
pack $textArea -side top -expand true -fill both
pack $textArea.yscroll -side right -fill y
pack $textArea.text -side left -fill both -expand true

set current_directory ""
set input_file_name ""
set fileChanged 0
set sgmlFormat 0
set num_sents 0



# Convert a text widget index (row.col) into an positive integer value
# Assume a max number of chars per line of 1024
proc IndexToInt {textIndex} {
    if { $textIndex != "" } {
	if {[regexp {^([0-9]+)\.([0-9]+)$} $textIndex match rowIndex colIndex]} {
	    return [expr $rowIndex*1024 + $colIndex]
	} else {
	    puts "IndexToInt: Bad argument: $textIndex"
	    return -1
	}
    } else {
	puts "IndexToInt: No argument."
	return -1
    }
}


# Return >0 if tag1 should come before tag2
#        <0 if tag2 should come before tag1
#        0  if tag1 and tag2 have the same end-points
# Assuming tags do not cross brackets
proc PreOrder { tag1 tag2 } {
    global rangeTable
    set from1 [IndexToInt [lindex $rangeTable($tag1) 0]]
    set to1   [IndexToInt [lindex $rangeTable($tag1) 1]]
    set from2 [IndexToInt [lindex $rangeTable($tag2) 0]]
    set to2   [IndexToInt [lindex $rangeTable($tag2) 1]]
    if {$from1 == $from2} {
	return [expr $to1 - $to2]
    }
    return [expr $from2 - $from1]
}


# Insert tag in a priorityList respecting the preorder sequence:
# Returns the updated list:
# Average worst case running time: O(List length = Average number of tag end-points per text index)
proc InsertInPriorityList {tag pList} {
    set size [llength $pList]
    set i 0
    set iPoint 0
    while {$i < $size} {
	if {[PreOrder [lindex $pList $i] $tag] <= 0} {
	    set i $size
	} else {
	    incr i
	    set iPoint $i
	}
    }
    return [linsert $pList $iPoint $tag]
}
	

# Load an sgml file and fills in sents, num_sents, listOfSgmlTags and rangeTable:
LoadFile $inputFile
wm title . "(N)\[b\] View Segments: [file tail $inputFile]"
wm iconname . "(N)\[b\] View Segments: [file tail $inputFile]"
wm title .outline "(N)\[b\] Segmentation Summary: [file tail $inputFile]"
wm iconname .outline "(N)\[b\] Segmentation Summary: [file tail $inputFile]"
wm minsize . 50 10 
set colorIndex 0

foreach oneTag $listOfSgmlTags {
    set tagHead $oneTag
    regexp {^([-a-zA-Z0-9]+)\_([0-9]+)$} $oneTag match tagHead value
    if {[string length $segmentType] > 1} {
	if {$tagHead == $segmentType} {
	    lappend segmentTags $oneTag
	}
    }
    if {[lsearch -exact $segmentTags $oneTag] != -1} {
	set fromIndex [lindex $rangeTable($oneTag) 0]
	set toIndex   [lindex $rangeTable($oneTag) 1]
	regexp {^([0-9]+)\.([0-9]+)} $fromIndex match fromRow fromCol
	regexp {^([0-9]+)\.([0-9]+)} $toIndex match toRow toCol
	lappend listOfSegmentTags $oneTag
	# Assign color:
	if {![info exists colorTable($oneTag)]} {
	    set colorTable($oneTag) [lindex $palette $colorIndex]
	    incr colorIndex
	    if {$colorIndex >= [llength $palette]} {
		set colorIndex 0
	    }
	}
	# Insert tag end-points in priority lists:
	set plist [list]
	if {[info exists beginAtRow($fromRow)]} {
	    set plist $beginAtRow($fromRow)
	}
	set beginAtRow($fromRow) [InsertInPriorityList $oneTag $plist]
	set plist [list]
	if {[info exists endAtRow($toRow)]} {
	    set plist $endAtRow($toRow)
	}
	set endAtRow($toRow) [InsertInPriorityList $oneTag $plist]
    }
}


set segments(0) [list TopLevel]
set colorTable(TopLevel) Black
set maxDepth 0
set oldRow 0
set outlineRow 0
for {set row 1} {$row < $num_sents} {incr row ; incr oldRow} {
    # Copy over all segments from row-1 that are still open:
    foreach oldSegment $segments($oldRow) {
	if {[info exists endAtRow($oldRow)]} {
	    if {[lsearch -exact $endAtRow($oldRow) $oldSegment] != -1} {
		# Do nothing
	    } else {
		lappend segments($row) $oldSegment
	    }
	} else {
	   lappend segments($row) $oldSegment
	} 
    }
    # Add the new segments that are open at new row:
    if {[info exists beginAtRow($row)]} {
	set segmentBuffer  $segments($row) 
	set segments($row) [concat $segments($row) $beginAtRow($row)]
	foreach seg $beginAtRow($row) {
	    incr outlineRow
	    set  outlineText($outlineRow) $idTable($seg)
	    set  outlineSeg($outlineRow)  [concat $segmentBuffer $seg]
	    set  segmentBuffer [concat $segmentBuffer $seg]
	}
    }
    # Update maxDepth:
    if {[llength $segments($row)] > $maxDepth} {
	set maxDepth [llength $segments($row)]
    }
}
for {set row 1} {$row < $num_sents} {incr row} {
    set realSegments($row) $segments($row)
    # Normalize segments($row) to maxDepth:
    while {[llength $segments($row)] < $maxDepth} {
	lappend segments($row) [lindex $segments($row) [expr [llength $segments($row)] - 1]]
    }
}

set tabs    "     "
set tabSize [string length $tabs]

for {set row 1} {$row < $num_sents} {incr row} {
    set col 0
    set spaces($row) ""
    set realspaces($row) ""
    # Indentation:
    foreach segment $segments($row) {
	set spaces($row) $spaces($row)$tabs
    }
    foreach segment $realSegments($row) {
	set realspaces($row) $realspaces($row)$tabs
    }
}


# Display text:
$textArea.text configure -state normal
$textArea.text delete 1.0 end
set index 1
while {$index < $num_sents} {
    set offset($index) [expr [string length $spaces($index)] + [string length $realspaces($index)]]
    set offset($index) [expr $offset($index) + 2]
    $textArea.text insert end [format "%s %s\n" $spaces($index) $sents($index)]
    incr index
}
$textArea.text configure -state disabled


# Display outline:
set maxwidth 0
.outline.1.text configure -state normal
.outline.1.text delete 1.0 end
.outline.1.text insert end [format " \n"]
set index 1
while {$index <= $outlineRow } {
    set outlineSpaces($index) ""
    set outlineRealSpaces($index) ""
    for {set i 1} {$i < [llength $outlineSeg($index)]} {incr i} {
	set outlineRealSpaces($index) $outlineRealSpaces($index)$tabs
    }
    while {[llength $outlineSeg($index)] < $maxDepth} {
	lappend outlineSeg($index) [lindex $outlineSeg($index) [expr [llength $outlineSeg($index)] - 1]]
    }
    foreach seg $outlineSeg($index) {
	set from [string length $outlineSpaces($index)]
	set outlineSpaces($index) $outlineSpaces($index)$tabs
	set to [string length $outlineSpaces($index)]
	set fromIndex [expr $index + 1].$from
	set toIndex [expr $index + 1].$to
	.outline.1.text tag add $segment $fromIndex $toIndex
	.outline.1.text tag configure $segment -foreground Black -background $colorTable($seg)
    }
    set str [format "%s %s %s \n" $outlineSpaces($index) $outlineRealSpaces($index)  $outlineText($index)]
    if {[string length $str] > $maxwidth} {
	set maxwidth [string length $str]
    }
    .outline.1.text insert end $str
    incr index
}
.outline.1.text insert end [format " \n"]


proc UniqueStrings { alist } {
    set result " "
    set spaces " "
    foreach elem $alist {
	if {[regexp " $elem " $result match value] == 1} {
	    set result $result
	} else {
	    set result $result$elem$spaces
	}
    }
    return $result
}


# Add colors:
for {set row 1} {$row < $num_sents} {incr row} {
    set col 0
    # Add colors for Indentation:
    foreach segment $segments($row) {
	set fromIndex "$row.$col"
	set col [expr [expr $col + $tabSize] - 1]
	set toIndex [$textArea.text index "$row.$col + 1 chars"]
	incr col
	$textArea.text tag add $segment $fromIndex $toIndex
	$textArea.text tag configure $segment -foreground Black -background $colorTable($segment)
	
    }
    # Add color for Text:
    # set fromIndex "$row.$col + $offset($row) chars"
    set fromIndex "$row.$col"
    set toIndex [$textArea.text index "$row.$col lineend"]
    # set toIndex [$textArea.text index "$row.$col + $offset($row) chars"]
    if {$colorTable($segment) != "Black"} {
	$textArea.text tag add text.$segment $fromIndex $toIndex
	$textArea.text tag configure text.$segment -foreground $colorTable($segment)
	set idTable(text.$segment) [UniqueStrings [lreplace $realSegments($row) 0 0]]
    }
    $textArea.text see $row.0
    update idletasks

}


set index 1
while {$index <= $outlineRow } {
    set outlineSpaces($index) ""
    foreach seg $outlineSeg($index) {
	set from [string length $outlineSpaces($index)]
	set outlineSpaces($index) $outlineSpaces($index)$tabs
	set to [string length $outlineSpaces($index)]
	set fromIndex [expr $index + 1].$from
	set toIndex [expr $index + 1].$to
	.outline.1.text tag add outline.$seg $fromIndex $toIndex
	
	.outline.1.text tag configure outline.$seg -foreground Black -background $colorTable($seg)
	
    }
    set seg [lindex $outlineSeg($index) [expr [llength $outlineSeg($index)] - 1]]
    set fromIndex $toIndex
    set str [format "%s %s %s \n" $outlineSpaces($index) $outlineRealSpaces($index)  $outlineText($index)]
    set to [expr $to + [string length $str]]
    set toIndex [expr $index + 1].$to
    .outline.1.text tag add text.$seg $fromIndex $toIndex
    .outline.1.text tag configure text.$seg -foreground $colorTable($seg) -background Black
    .outline.1.text see $fromIndex
    update idletasks
    incr index
}

set h [expr $outlineRow + 2]
if {$h > 25} {
    set h 25
}

if {$maxwidth > 80} {
    set maxwidth 80
}

.outline.1.text configure -state disabled -height $h -width $maxwidth


# List all the current idTable(Tags) under the cursor:
proc ListTagsUnderCursor {} {
    global idTable tagUnderCursor textArea
    set listId ""
    set spaces " "
    foreach oneTag [$textArea.text tag names "current"] {
	if {[info exists idTable($oneTag)]} {
	    if {[regexp {( $idTable($oneTag))} $listId match value] == 1} {
	        # set listId $listId 
	    } else {
		set listId $listId$spaces$idTable($oneTag)
	    }
	}
    }
    set tagUnderCursor $listId
}

# Bind all segment tags to show the segment under the cursor:
if {[info exists listOfSegmentTags]} {
    if {[llength $listOfSegmentTags] > 0} {
	foreach newTag $listOfSegmentTags {
	    $textArea.text tag bind $newTag <Enter> {
		ListTagsUnderCursor
		break
	    }
	    $textArea.text tag bind $newTag <Leave> {
		set tagUnderCursor ""
		break
	    }
	}
    }
}


pack .outline.1 -side top -fill both -expand true
pack .outline.1.yscroll -side right -fill y 
pack .outline.1.text -side left -fill both -expand true

pack .outline.2 -fill x -expand false
pack .outline.2.pad -side right
pack .outline.2.xscroll -side left -expand true -fill x

pack .outline.3 -fill x -expand false
pack .outline.3.pad -side right
pack .outline.3.dismiss -side right


